"""Submodule containing the functions relating to grid points."""

import logging

log = logging.getLogger(__name__)

import numpy as np
import numpy.ma as ma

import resqpy.olio.point_inclusion as pip
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec
import resqpy.olio.xml_et as rqet
import resqpy.property as rprop
from ._defined_geometry import pillar_geometry_is_defined, cell_geometry_is_defined, geometry_defined_for_all_cells, \
    geometry_defined_for_all_pillars


def set_cached_points_from_property(grid,
                                    points_property_uuid = None,
                                    property_collection = None,
                                    realization = None,
                                    time_index = None,
                                    set_grid_time_index = True,
                                    set_inactive = True,
                                    active_property_uuid = None,
                                    active_collection = None):
    """Modifies the cached points (geometry), setting the values from a points property.

    arguments:
       points_property_uuid (uuid, optional): the uuid of the points property; if present the
          remaining arguments are ignored except for inactive & active arguments
       property_collection (PropertyCollection, optional): defaults to property collection
          for the grid; should only contain one set of points properties (but may also contain
          other non-points properties)
       realization (int, optional): if present, the property in the collection with this
          realization number is used
       time_index (int, optional): if present, the property in the collection with this
          time index is used
       set_grid_time_index (bool, default True): if True, the grid's time index will be set
          to the time_index argument and the grid's time series uuid will be set to that
          referred to by the points property; if False, the grid's time index will not be
          modified
       set_inactive (bool, default True): if True, the grid's inactive mask will be set
          based on an active cell property
       active_property_uuid (uuid, optional): if present, the uuid of an active cell property
          to base the inactive mask on; ignored if set_inactive is False
       active_collection (uuid, optional): default's to property_collection if present, or
          the grid's property collection otherwise; only used if set_inactive is True and
          active_property_uuid is None

    notes:
       the points property must have indexable element 'nodes' and be the same shape as the
       official points array for the grid;
       note that the shape of the points array is quite different between grids with split
       pillars and those without;
       the uom of the points property must be a length uom and match that used by the grid's crs;
       the inactive mask of the grid will only be updated if the set_inactive argument is True;
       if points_property_uuid has been provided, and set_inactive is True, the active property
       must be identified with the active_property_uuid argument;
       if set_inactive is True and active_property_uuid is None and points_property_uuid is None and
       realization and/or time_index is in use, the active property collection must contain one
       series of active properties with the same variants (realizations and time indices) as the
       points property series;
       the active cell properties should be discrete and have a local property kind titled 'active';
       various cached data are invalidated and cleared by this method
    """

    if points_property_uuid is None:
        if property_collection is None:
            property_collection = grid.extract_property_collection()
        part = property_collection.singleton(points = True,
                                             indexable = 'nodes',
                                             realization = realization,
                                             time_index = time_index)
        assert part is not None, 'failed to identify points property to use for grid geometry'
        points_property_uuid = property_collection.uuid_for_part(part)
    elif set_inactive:
        assert active_property_uuid is not None, 'active property uuid not provided when setting points from property'

    assert points_property_uuid is not None

    grid.cache_all_geometry_arrays()  # the split pillar information must not vary

    # check for compatibility and overwrite cached points for grid
    points = rprop.Property(grid.model, uuid = points_property_uuid)
    assert points is not None and points.is_points() and points.indexable_element() == 'nodes'
    assert points.uom() == grid.xy_units() and grid.z_units() == grid.xy_units()
    points_array = points.array_ref(masked = False)
    assert points_array is not None
    assert points_array.shape == grid.points_cached.shape
    grid.points_cached = points_array

    # set grid's time index and series, if requested
    if set_grid_time_index:
        grid.time_index = points.time_index()
        ts_uuid = points.time_series_uuid()
        if ts_uuid is not None and grid.time_series_uuid is not None:
            if not bu.matching_uuids(ts_uuid, grid.time_series_uuid):
                log.warning('change of time series uuid for dynamic grid geometry')
        grid.time_series_uuid = ts_uuid

    # invalidate anything cached that is derived from geometry
    grid.geometry_defined_for_all_pillars_cached = None
    grid.geometry_defined_for_all_cells_cached = None
    for attr in ('array_unsplit_points', 'array_corner_points', 'array_centre_point', 'array_thickness', 'array_volume',
                 'array_half_cell_t', 'array_k_transmissibility', 'array_j_transmissibility',
                 'array_i_transmissibility', 'fgcs', 'array_fgcs_transmissibility', 'pgcs',
                 'array_pgcs_transmissibility', 'kgcs', 'array_kgcs_transmissibility'):
        if hasattr(grid, attr):
            delattr(grid, attr)

    if set_inactive:
        if active_property_uuid is None:
            if active_collection is None:
                active_collection = property_collection
                assert active_collection is not None
            active_part = active_collection.singleton(property_kind = 'active',
                                                      indexable = 'cells',
                                                      continuous = False,
                                                      realization = realization,
                                                      time_index = time_index)
            assert active_part is not None, 'failed to identify active property to use for grid inactive mask'
            active_property_uuid = active_collection.uuid_for_part(active_part)
        active = rprop.Property(grid.model, uuid = active_property_uuid)
        assert active is not None
        active_array = active.array_ref()
        assert active_array.shape == tuple(grid.extent_kji)
        grid.inactive = np.logical_not(active_array)
        grid.all_inactive = np.all(grid.inactive)
        grid.active_property_uuid = active_property_uuid


def point_raw(grid, index = None, points_root = None, cache_array = True):
    """Returns element from points data, indexed as in the hdf5 file; can optionally be used to cache points data.

    arguments:
       index (2 or 3 integers, optional): if not None, the index into the raw points data for the point of interest
       points_root (optional): the xml node holding the points data
       cache_array (boolean, default True): if True, the raw points data is cached in memory as a side effect

    returns:
       (x, y, z) of selected point as a 3 element numpy vector, or None if index is None

    notes:
       this function is typically called either to cache the points data in memory, or to fetch the coordinates of
       a single point; the details of the indexing depend upon whether the grid has split coordinate lines: if not,
       the index should be a triple kji0 with axes ranging over the shared corners nk+k_gaps+1, nj+1, ni+1; if there
       are split pillars, index should be a pair, being the k0 in range nk+k_gaps+1 and a pillar index; note that if
       index is passed, the k0 element must already have been mapped to the raw index value taking into consideration
       any k gaps; if the grid object does not include geometry then None is returned
    """

    # NB: shape of index depends on whether grid has split pillars
    if index is not None and not geometry_defined_for_all_pillars(grid, cache_array = cache_array):
        if len(index) == 3:
            ji = tuple(index[1:])
        else:
            ji = tuple(divmod(index[1], grid.ni))
        if ji[0] < grid.nj and not pillar_geometry_is_defined(grid, ji, cache_array = cache_array):
            return None
    if grid.points_cached is not None:
        if index is None:
            return grid.points_cached
        return grid.points_cached[tuple(index)]
    p_root = grid.resolve_geometry_child('Points', child_node = points_root)
    if p_root is None:
        log.debug('point_raw() returning None as geometry not present')
        return None  # geometry not present
    assert rqet.node_type(p_root) == 'Point3dHdf5Array'
    h5_key_pair = grid.model.h5_uuid_and_path_for_node(p_root, tag = 'Coordinates')
    if h5_key_pair is None:
        return None
    if grid.has_split_coordinate_lines:
        required_shape = None
    else:
        required_shape = (grid.nk_plus_k_gaps + 1, grid.nj + 1, grid.ni + 1, 3)
    try:
        value = grid.model.h5_array_element(h5_key_pair,
                                            index = index,
                                            cache_array = cache_array,
                                            object = grid,
                                            array_attribute = 'points_cached',
                                            required_shape = required_shape)
    except Exception:
        log.error('hdf5 points failure for index: ' + str(index))
        raise
    if index is None:
        return grid.points_cached
    return value


def point(grid, cell_kji0 = None, corner_index = np.zeros(3, dtype = 'int'), points_root = None, cache_array = True):
    """Return a cell corner point xyz; can optionally be used to cache points data.

    arguments:
       cell_kji0 (3 integers, optional): if not None, the index of the cell for the point of interest, in kji0 protocol
       corner_index (3 integers, default zeros): the kp, jp, ip corner-within-cell indices (each 0 or 1)
       points_root (optional): the xml node holding the points data
       cache_array (boolean, default True): if True, the raw points data is cached in memory as a side effect

    returns:
       (x, y, z) of selected point as a 3 element numpy vector, or None if cell_kji0 is None

    note:
       if cell_kji0 is passed, the k0 value should be the layer index before adjustment for k_gaps, which this
       method will apply
    """

    if cache_array and grid.points_cached is None:
        grid.point_raw(points_root = points_root, cache_array = True)
    if cell_kji0 is None:
        return None
    if grid.k_raw_index_array is None:
        grid.extract_k_gaps()
    if not geometry_defined_for_all_cells(grid):
        if not cell_geometry_is_defined(grid, cell_kji0 = cell_kji0, cache_array = cache_array):
            return None
    p_root = grid.resolve_geometry_child('Points', child_node = points_root)
    #      if p_root is None: return None  # geometry not present
    index = np.zeros(3, dtype = int)
    index[:] = cell_kji0
    index[0] = grid.k_raw_index_array[index[0]]  # adjust for k gaps
    if grid.has_split_coordinate_lines:
        grid.create_column_pillar_mapping()
        pillar_index = grid.pillars_for_column[index[1], index[2], corner_index[1], corner_index[2]]
        return grid.point_raw(index = (index[0] + corner_index[0], pillar_index),
                              points_root = p_root,
                              cache_array = cache_array)
    else:
        index[:] += corner_index
        return grid.point_raw(index = index, points_root = p_root, cache_array = cache_array)


def points_ref(grid, masked = True):
    """Returns an in-memory numpy array containing the xyz data for points used in the grid geometry.

    argument:
       masked (boolean, default True): if True, a masked array is returned with NaN points masked out;
          if False, a simple (unmasked) numpy array is returned

    returns:
       numpy array or masked array of float, of shape (nk + k_gaps + 1, nj + 1, ni + 1, 3) or (nk + k_gaps + 1, np, 3)
       where np is the total number of pillars (primary pillars + extras for split pillars)

    notes:
       this is the usual way to get at the actual grid geometry points data in the native resqml layout;
       the has_split_coordinate_lines boolean attribute can be used to determine which shape to expect;
       the shape is (nk + k_gaps + 1, nj + 1, ni + 1, 3) if there are no split coordinate lines (unfaulted);
       otherwise it is (nk + k_gaps + 1, np, 3), where np > (nj + 1) * (ni + 1), due to extra pillar data for
       the split pillars

    :meta common:
    """

    if grid.points_cached is None:
        grid.point_raw(cache_array = True)
        if grid.points_cached is None:
            return None
    if not masked:
        return grid.points_cached
    return ma.masked_invalid(grid.points_cached)


def uncache_points(grid):
    """Frees up memory by removing the cached copy of the grid's points data.

    note:
       the memory will only actually become free when any other references to it pass out of scope
       or are deleted
    """

    if grid.points_cached is not None:
        del grid.points_cached
        grid.points_cached = None


def unsplit_points_ref(grid, cache_array = False, masked = False):
    """Returns a copy of the points array that has split pillars merged back into an unsplit configuration.

    arguments:
       cache_array (boolean, default False): if True, a copy of the unsplit points array is added as
          attribute array_unsplit_points to this grid object
       masked (boolean, default False): if True, a masked array is returned with NaN points masked out;
          if False, a simple (unmasked) numpy array is returned

    returns:
       numpy array of float of shape (nk + k_gaps + 1, nj + 1, ni + 1, 3)

    note:
       for grids without split pillars, this function simply returns the points array in its native form;
       for grids with split pillars, an unsplit equivalent points array is calculated as the average of
       contributions to each pillar from the surrounding cell columns
    """

    if hasattr(grid, 'array_unsplit_points'):
        return grid.array_unsplit_points
    points = grid.points_ref(masked = masked)
    if not grid.has_split_coordinate_lines:
        if cache_array:
            grid.array_unsplit_points = points.copy()
            return grid.array_unsplit_points
        return points
    # todo: finish version that copies primaries and only modifies split pillars?
    # njkp1 = (grid.nj + 1) * (grid.ni + 1)
    # merged_points = np.empty((grid.nk + 1, njkp1, 3))   # shaped somewhat like split points array
    # merged_points[:, :, :] = points[:, :njkp1, :]       # copy primary data
    result = np.empty((grid.nk_plus_k_gaps + 1, grid.nj + 1, grid.ni + 1, 3))
    # todo: if not geometry defined for all cells, take nanmean of four points?
    # compute for internal pillars
    grid.create_column_pillar_mapping()  # pillar indices for 4 columns around interior pillars
    pfc_11 = grid.pillars_for_column[:-1, :-1, 1, 1]
    pfc_10 = grid.pillars_for_column[:-1, 1:, 1, 0]
    pfc_01 = grid.pillars_for_column[1:, :-1, 0, 1]
    pfc_00 = grid.pillars_for_column[1:, 1:, 0, 0]
    result[:, 1:-1,
           1:-1, :] = 0.25 * (points[:, pfc_11, :] + points[:, pfc_10, :] + points[:, pfc_01, :] + points[:, pfc_00, :])
    # edges
    # todo: use numpy array operations instead of for loops (see lines above for example code)
    for j in range(1, grid.nj):
        result[:, j, 0, :] = 0.5 * (points[:, grid.pillars_for_column[j - 1, 0, 1, 0], :] +
                                    points[:, grid.pillars_for_column[j, 0, 0, 0], :])
        result[:, j, grid.ni, :] = 0.5 * (points[:, grid.pillars_for_column[j - 1, grid.ni - 1, 1, 1], :] +
                                          points[:, grid.pillars_for_column[j, grid.ni - 1, 0, 1], :])
    for i in range(1, grid.ni):
        result[:, 0, i, :] = 0.5 * (points[:, grid.pillars_for_column[0, i - 1, 0, 1], :] +
                                    points[:, grid.pillars_for_column[0, i, 0, 0], :])
        result[:, grid.nj, i, :] = 0.5 * (points[:, grid.pillars_for_column[grid.nj - 1, i - 1, 1, 1], :] +
                                          points[:, grid.pillars_for_column[grid.nj - 1, i, 1, 0], :])
    # corners (could optimise as these should always be primaries
    result[:, 0, 0, :] = points[:, grid.pillars_for_column[0, 0, 0, 0], :]
    result[:, 0, grid.ni, :] = points[:, grid.pillars_for_column[0, grid.ni - 1, 0, 1], :]
    result[:, grid.nj, 0, :] = points[:, grid.pillars_for_column[grid.nj - 1, 0, 1, 0], :]
    result[:, grid.nj, grid.ni, :] = points[:, grid.pillars_for_column[grid.nj - 1, grid.ni - 1, 1, 1], :]
    if cache_array:
        grid.array_unsplit_points = result
        return grid.array_unsplit_points
    return result


def horizon_points(grid, ref_k0 = 0, heal_faults = False, kp = 0):
    """Returns reference to a points layer array of shape ((nj + 1), (ni + 1), 3) based on primary pillars.

    arguments:
       ref_k0 (integer): the horizon layer number, in the range 0 to nk (or layer number in range 0..nk-1
          in the case of grids with k gaps)
       heal_faults (boolean, default False): if True and the grid has split coordinate lines, an unsplit
          equivalent of the grid points is generated first and the returned points are based on that data;
          otherwise, the primary pillar data is used, which effectively gives a point from one side or
          another of any faults, rather than an averaged point
       kp (integer, default 0): set to 1 to specify the base of layer ref_k0, in case of grids with k gaps

    returns:
       a numpy array of floats of shape ((nj + 1), (ni + 1), 3) being the (shared) cell corner point
       locations for the plane of points, based on the primary pillars or unsplit equivalent pillars

    notes:
       the primary pillars are the 'first' set of points for a pillar; a split pillar will have one to
       three other sets of point data but those are ignored by this function unless heal_faults is True,
       in which case an averaged point will be used for the split pillars;
       to get full unhealed representation of split horizon points, use split_horizon_points() function
       instead;
       for grids without k gaps, ref_k0 can be used alone, in the range 0..nk, to identify the horizon;
       alternatively, or for grids with k gaps, the ref_k0 can specify the layer in the range 0..nk-1,
       with kp being passed the value 0 (default) for the top of the layer, or 1 for the base of the layer
    """

    # note: if heal_faults is False, primary pillars only are used
    pe_j = grid.nj + 1
    pe_i = grid.ni + 1
    if grid.k_gaps:
        ref_k0 = grid.k_raw_index_array[ref_k0]
    ref_k0 += kp
    if grid.has_split_coordinate_lines:
        if heal_faults:
            points = grid.unsplit_points_ref()  # expensive operation: would be better to cache the unsplit points
            return points[ref_k0, :, :, :].reshape((pe_j, pe_i, 3))
        else:
            points = grid.points_ref(masked = False)
            return points[ref_k0, :pe_j * pe_i, :].reshape((pe_j, pe_i, 3))
    # unfaulted grid
    points = grid.points_ref(masked = False)
    return points[ref_k0, :, :, :].reshape((pe_j, pe_i, 3))


def split_horizon_points(grid, ref_k0 = 0, masked = False, kp = 0):
    """Returns reference to a corner points for a horizon, of shape (nj, ni, 2, 2, 3).

    arguments:
       ref_k0 (integer): the horizon layer number, in the range 0 to nk (or layer number in range 0..nk-1
          in the case of grids with k gaps)
       masked (boolean, default False): if True, a masked array is returned with NaN points masked out;
          if False, a simple (unmasked) numpy array is returned
       kp (integer, default 0): set to 1 to specify the base of layer ref_k0, in case of grids with k gaps

    returns:
       numpy array of shape (nj, ni, 2, 2, 3) being corner point x,y,z values for cell corners (j, i, jp, ip)

    notes:
       if split points are needed for a range of horizons, it is more efficient to call split_horizons_points()
       than repeatedly call this function;
       for grids without k gaps, ref_k0 can be used alone, in the range 0..nk, to identify the horizon;
       alternatively, or for grids with k gaps, the ref_k0 can specify the layer in the range 0..nk-1,
       with kp being passed the value 0 (default) for the top of the layer, or 1 for the base of the layer
    """

    if grid.k_gaps:
        ref_k0 = grid.k_raw_index_array[ref_k0]
    ref_k0 += kp
    points = grid.points_ref(masked = masked)
    hp = np.empty((grid.nj, grid.ni, 2, 2, 3))
    if grid.has_split_coordinate_lines:
        assert points.ndim == 3
        # todo: replace for loops with numpy slice indexing
        grid.create_column_pillar_mapping()
        assert grid.pillars_for_column.ndim == 4 and grid.pillars_for_column.shape == (grid.nj, grid.ni, 2, 2)
        for j in range(grid.nj):
            for i in range(grid.ni):
                hp[j, i, 0, 0, :] = points[ref_k0, grid.pillars_for_column[j, i, 0, 0], :]
                hp[j, i, 1, 0, :] = points[ref_k0, grid.pillars_for_column[j, i, 1, 0], :]
                hp[j, i, 0, 1, :] = points[ref_k0, grid.pillars_for_column[j, i, 0, 1], :]
                hp[j, i, 1, 1, :] = points[ref_k0, grid.pillars_for_column[j, i, 1, 1], :]
    else:
        assert points.ndim == 4
        hp[:, :, 0, 0, :] = points[ref_k0, :-1, :-1, :]
        hp[:, :, 1, 0, :] = points[ref_k0, 1:, :-1, :]
        hp[:, :, 0, 1, :] = points[ref_k0, :-1, 1:, :]
        hp[:, :, 1, 1, :] = points[ref_k0, 1:, 1:, :]
    return hp


def split_x_section_points(grid, axis, ref_slice0 = 0, plus_face = False, masked = False):
    """Returns an array of points representing cell corners from an I or J interface slice for a faulted grid.

    arguments:
       axis (string): 'I' or 'J' being the axis of the cross-sectional slice (ie. dimension being dropped)
       ref_slice0 (int, default 0): the reference value for indices in I or J (as defined in axis)
       plus_face (boolean, default False): if False, negative face is used; if True, positive
       masked (boolean, default False): if True, a masked numpy array is returned with NaN values masked out

    returns:
       a numpy array of shape (nk + 1, nj, 2, 3) or (nk + 1, ni, 2, 3) being the xyz points of the cell corners
       on the interfacial cross section; 3rd axis is jp or ip; final axis is xyz

    note:
       this function will only work for grids with no k gaps; it is intended for split grids though will also
       function for unsplit grids; use split_gap_x_section_points() if k gaps are present
    """

    log.debug(f'x-sect: axis {axis}; ref_slice0 {ref_slice0}; plus_face {plus_face}; masked {masked}')
    assert axis.upper() in ['I', 'J']
    assert not grid.k_gaps, 'split_x_section_points() method is for grids without k gaps; use split_gap_x_section_points()'

    points = grid.points_ref(masked = masked)
    cpm = grid.create_column_pillar_mapping()

    ij_p = 1 if plus_face else 0

    if axis.upper() == 'I':
        return points[:, cpm[:, ref_slice0, :, ij_p], :]
    else:
        return points[:, cpm[ref_slice0, :, ij_p, :], :]


def split_gap_x_section_points(grid, axis, ref_slice0 = 0, plus_face = False, masked = False):
    """Return array of points representing cell corners from an I or J interface slice for a faulted grid.

    arguments:
       axis (string): 'I' or 'J' being the axis of the cross-sectional slice (ie. dimension being dropped)
       ref_slice0 (int, default 0): the reference value for indices in I or J (as defined in axis)
       plus_face (boolean, default False): if False, negative face is used; if True, positive
       masked (boolean, default False): if True, a masked numpy array is returned with NaN values masked out

    returns:
       a numpy array of shape (nk, nj, 2, 2, 3) or (nk, ni, 2, 2, 3) being the xyz points of the cell corners
       on the interfacial cross section; 3rd axis is kp; 4th axis is jp or ip; final axis is xyz

    note:
       this function is intended for split grids with k gaps though will also function for split grids
       without k gaps
    """

    log.debug(f'k gap x-sect: axis {axis}; ref_slice0 {ref_slice0}; plus_face {plus_face}; masked {masked}')
    assert axis.upper() in ['I', 'J']

    if grid.has_split_coordinate_lines:
        points = grid.points_ref(masked = masked)
    else:
        points = grid.points_ref(masked = masked).reshape((grid.nk_plus_k_gaps, (grid.nj + 1) * (grid.ni + 1), 3))
    cpm = grid.create_column_pillar_mapping()

    ij_p = 1 if plus_face else 0

    if grid.k_gaps:
        top_points = points[grid.k_raw_index_array]
        base_points = points[grid.k_raw_index_array + 1]
        if axis.upper() == 'I':
            top = top_points[:, cpm[:, ref_slice0, :, ij_p], :]
            base = base_points[:, cpm[:, ref_slice0, :, ij_p], :]
        else:
            top = top_points[:, cpm[ref_slice0, :, ij_p, :], :]
            base = base_points[:, cpm[ref_slice0, :, ij_p, :], :]
    else:
        p = grid.split_x_section_points(axis, ref_slice0 = ref_slice0, plus_face = plus_face, masked = masked)
        top = p[:-1]
        base = p[1:]
    return np.stack((top, base), axis = 2)


def unsplit_x_section_points(grid, axis, ref_slice0 = 0, plus_face = False, masked = False):
    """Returns a 2D (+1 for xyz) array of points representing cell corners from an I or J interface slice.

    arguments:
       axis (string): 'I' or 'J' being the axis of the cross-sectional slice (ie. dimension being dropped)
       ref_slice0 (int, default 0): the reference value for indices in I or J (as defined in axis)
       plus_face (boolean, default False): if False, negative face is used; if True, positive
       masked (boolean, default False): if True, a masked numpy array is returned with NaN values masked out

    returns:
       a 2+1D numpy array being the xyz points of the cell corners on the interfacial cross section;
       the 2D axes are K,J or K,I - whichever does not involve axis; shape is (nk + 1, nj + 1, 3) or
       (nk + 1, ni + 1, 3)

    note:
       restricted to unsplit grids with no k gaps; use split_x_section_points() for split grids with no k gaps
       or split_gap_x_section_points() for split grids with k gaps or x_section_corner_points() for any grid
    """

    log.debug(f'x-sect: axis {axis}; ref_slice0 {ref_slice0}; plus_face {plus_face}; masked {masked}')
    assert axis.upper() in ['I', 'J']
    assert not grid.has_split_coordinate_lines, 'cross sectional points for unsplit grids require split_x_section_points()'
    assert not grid.k_gaps, 'cross sectional points with k gaps require split_gap_x_section_points()'

    if plus_face:
        ref_slice0 += 1

    points = grid.points_ref(masked = masked)

    if axis.upper() == 'I':
        return points[:, :, ref_slice0, :]
    else:
        return points[:, ref_slice0, :, :]


def x_section_corner_points(grid,
                            axis,
                            ref_slice0 = 0,
                            plus_face = False,
                            masked = False,
                            rotate = False,
                            azimuth = None):
    """Returns a fully expanded array of points representing cell corners from an I or J interface slice.

    arguments:
       axis (string): 'I' or 'J' being the axis of the cross-sectional slice (ie. dimension being dropped)
       ref_slice0 (int, default 0): the reference value for indices in I or J (as defined in axis)
       plus_face (boolean, default False): if False, negative face is used; if True, positive
       masked (boolean, default False): if True, a masked numpy array is returned with NaN values masked out
       rotate (boolean, default False): if True, the cross section points are rotated around the z axis so that
          an azimuthal direction is mapped onto the positive x axis
       aximuth (float, optional): the compass bearing in degrees to map onto the positive x axis if rotating;
          if None, the mean direction of the cross sectional points, along axis, is used; ignored if rotate
          is False

    returns:
       a numpy float array of shape (nk, nj, 2, 2, 3) or (nk, ni, 2, 2, 3) being the xyz points of the cell
       corners on the interfacial cross section; the 3rd index (1st 2) is kp, the 4th index is jp or ip

    note:
       this method will work for unsplit or split grids, with or without k gaps; use rotate argument to yield
       points with predominant variation in xz, suitable for plotting cross sections; if rotate is True then
       the absolute values of x & y will not be very meaningful though the units will still be the grid's xy
       units for relative purposes
    """

    assert axis.upper() in ['I', 'J']
    nj_or_ni = grid.nj if axis.upper() == 'I' else grid.ni

    if grid.k_gaps:
        x_sect = grid.split_gap_x_section_points(axis, ref_slice0 = ref_slice0, plus_face = plus_face, masked = masked)
    else:
        if grid.has_split_coordinate_lines:
            no_k_gap_xs = grid.split_x_section_points(axis,
                                                      ref_slice0 = ref_slice0,
                                                      plus_face = plus_face,
                                                      masked = masked)
            x_sect = np.empty((grid.nk, nj_or_ni, 2, 2, 3))
            x_sect[:, :, 0, :, :] = no_k_gap_xs[:-1, :, :, :]
            x_sect[:, :, 1, :, :] = no_k_gap_xs[1:, :, :, :]
        else:
            simple_xs = grid.unsplit_x_section_points(axis,
                                                      ref_slice0 = ref_slice0,
                                                      plus_face = plus_face,
                                                      masked = masked)
            x_sect = np.empty((grid.nk, nj_or_ni, 2, 2, 3))
            x_sect[:, :, 0, 0, :] = simple_xs[:-1, :-1, :]
            x_sect[:, :, 1, 0, :] = simple_xs[1:, :-1, :]
            x_sect[:, :, 0, 1, :] = simple_xs[:-1, 1:, :]
            x_sect[:, :, 1, 1, :] = simple_xs[1:, 1:, :]

    if rotate:
        if azimuth is None:
            direction = vec.points_direction_vector(x_sect, axis = 1)
        else:
            direction = vec.unit_vector_from_azimuth(azimuth)
        x_sect = vec.rotate_xyz_array_around_z_axis(x_sect, direction)
        x_sect[..., 0] -= np.nanmin(x_sect[..., 0])
        x_sect[..., 1] -= np.nanmin(x_sect[..., 1])

    return x_sect


def coordinate_line_end_points(grid):
    """Returns xyz of top and bottom of each primary pillar.

    returns:
       numpy float array of shape (nj + 1, ni + 1, 2, 3)
    """

    points = grid.points_ref(masked = False).reshape((grid.nk + 1, -1, 3))
    primary_pillar_count = (grid.nj + 1) * (grid.ni + 1)
    result = np.empty((grid.nj + 1, grid.ni + 1, 2, 3))
    result[:, :, 0, :] = points[0, :primary_pillar_count, :].reshape((grid.nj + 1, grid.ni + 1, 3))
    result[:, :, 1, :] = points[-1, :primary_pillar_count, :].reshape((grid.nj + 1, grid.ni + 1, 3))
    return result


def z_corner_point_depths(grid, order = 'cellular'):
    """Returns the z (depth) values of each corner of each cell.

    arguments:
       order (string, default 'cellular'): either 'cellular' or 'linear'; if 'cellular' the resulting array has
          shape (nk, nj, ni, 2, 2, 2); if 'linear', the shape is (nk, 2, nj, 2, ni, 2)

    returns:
       numpy array of shape (nk, nj, ni, 2, 2, 2) or (nk, 2, nj, 2, ni, 2); for the cellular ordering, the
       result can be indexed with [k, j, i, kp, jp, ip] (where kp, for example, is 0 for the K- face and 1 for K+);
       for the linear ordering, the equivalent indexing is [k, kp, j, jp, i, ip], as used by some common simulator
       keyword formats
    """

    assert order in ['cellular', 'linear']

    z_cp = np.empty((grid.nk, grid.nj, grid.ni, 2, 2, 2))
    points = grid.points_ref()
    if grid.has_split_coordinate_lines:
        grid.create_column_pillar_mapping()
        # todo: replace j,i for loops with numpy broadcasting
        if grid.k_gaps:
            for j in range(grid.nj):
                for i in range(grid.ni):
                    z_cp[:, j, i, 0, 0, 0] = points[grid.k_raw_index_array, grid.pillars_for_column[j, i, 0, 0], 2]
                    z_cp[:, j, i, 1, 0, 0] = points[grid.k_raw_index_array + 1, grid.pillars_for_column[j, i, 0, 0], 2]
                    z_cp[:, j, i, 0, 1, 0] = points[grid.k_raw_index_array, grid.pillars_for_column[j, i, 1, 0], 2]
                    z_cp[:, j, i, 1, 1, 0] = points[grid.k_raw_index_array + 1, grid.pillars_for_column[j, i, 1, 0], 2]
                    z_cp[:, j, i, 0, 0, 1] = points[grid.k_raw_index_array, grid.pillars_for_column[j, i, 0, 1], 2]
                    z_cp[:, j, i, 1, 0, 1] = points[grid.k_raw_index_array + 1, grid.pillars_for_column[j, i, 0, 1], 2]
                    z_cp[:, j, i, 0, 1, 1] = points[grid.k_raw_index_array, grid.pillars_for_column[j, i, 1, 1], 2]
                    z_cp[:, j, i, 1, 1, 1] = points[grid.k_raw_index_array + 1, grid.pillars_for_column[j, i, 1, 1], 2]
        else:
            for j in range(grid.nj):
                for i in range(grid.ni):
                    z_cp[:, j, i, 0, 0, 0] = points[:-1, grid.pillars_for_column[j, i, 0, 0], 2]
                    z_cp[:, j, i, 1, 0, 0] = points[1:, grid.pillars_for_column[j, i, 0, 0], 2]
                    z_cp[:, j, i, 0, 1, 0] = points[:-1, grid.pillars_for_column[j, i, 1, 0], 2]
                    z_cp[:, j, i, 1, 1, 0] = points[1:, grid.pillars_for_column[j, i, 1, 0], 2]
                    z_cp[:, j, i, 0, 0, 1] = points[:-1, grid.pillars_for_column[j, i, 0, 1], 2]
                    z_cp[:, j, i, 1, 0, 1] = points[1:, grid.pillars_for_column[j, i, 0, 1], 2]
                    z_cp[:, j, i, 0, 1, 1] = points[:-1, grid.pillars_for_column[j, i, 1, 1], 2]
                    z_cp[:, j, i, 1, 1, 1] = points[1:, grid.pillars_for_column[j, i, 1, 1], 2]
    else:
        if grid.k_gaps:
            z_cp[:, :, :, 0, 0, 0] = points[grid.k_raw_index_array, :-1, :-1, 2]
            z_cp[:, :, :, 1, 0, 0] = points[grid.k_raw_index_array + 1, :-1, :-1, 2]
            z_cp[:, :, :, 0, 1, 0] = points[grid.k_raw_index_array, 1:, :-1, 2]
            z_cp[:, :, :, 1, 1, 0] = points[grid.k_raw_index_array + 1, 1:, :-1, 2]
            z_cp[:, :, :, 0, 0, 1] = points[grid.k_raw_index_array, :-1, 1:, 2]
            z_cp[:, :, :, 1, 0, 1] = points[grid.k_raw_index_array + 1, :-1, 1:, 2]
            z_cp[:, :, :, 0, 1, 1] = points[grid.k_raw_index_array, 1:, 1:, 2]
            z_cp[:, :, :, 1, 1, 1] = points[grid.k_raw_index_array + 1, 1:, 1:, 2]
        else:
            z_cp[:, :, :, 0, 0, 0] = points[:-1, :-1, :-1, 2]
            z_cp[:, :, :, 1, 0, 0] = points[1:, :-1, :-1, 2]
            z_cp[:, :, :, 0, 1, 0] = points[:-1, 1:, :-1, 2]
            z_cp[:, :, :, 1, 1, 0] = points[1:, 1:, :-1, 2]
            z_cp[:, :, :, 0, 0, 1] = points[:-1, :-1, 1:, 2]
            z_cp[:, :, :, 1, 0, 1] = points[1:, :-1, 1:, 2]
            z_cp[:, :, :, 0, 1, 1] = points[:-1, 1:, 1:, 2]
            z_cp[:, :, :, 1, 1, 1] = points[1:, 1:, 1:, 2]

    if order == 'linear':
        return np.transpose(z_cp, axes = (0, 3, 1, 4, 2, 5))
    return z_cp


def corner_points(grid, cell_kji0 = None, points_root = None, cache_resqml_array = True, cache_cp_array = False):
    """Returns a numpy array of corner points for a single cell or the whole grid.

    notes:
       if cell_kji0 is not None, a 4D array of shape (2, 2, 2, 3) holding single cell corner points in logical order
       [kp, jp, ip, xyz] is returned; if cell_kji0 is None, a pagoda style 7D array [k, j, i, kp, jp, ip, xyz] is
       cached and returned;
       the ordering of the corner points is in the logical order, which is not the same as that used by Nexus CORP data;
       olio.grid_functions.resequence_nexus_corp() can be used to switch back and forth between this pagoda ordering
       and Nexus corp ordering;
       this is the usual way to access full corner points for cells where working with native resqml data is undesirable

    :meta common:
    """

    # note: this function returns a derived object rather than a native resqml object

    def one_cell_cp(grid, cell_kji0, points_root, cache_array):
        cp = np.full((2, 2, 2, 3), np.NaN)
        if not grid.geometry_defined_for_all_cells():
            if not grid.cell_geometry_is_defined(cell_kji0, cache_array = cache_array):
                return cp
        corner_index = np.zeros(3, dtype = 'int')
        for kp in range(2):
            corner_index[0] = kp
            for jp in range(2):
                corner_index[1] = jp
                for ip in range(2):
                    corner_index[2] = ip
                    one_point = grid.point(cell_kji0,
                                           corner_index = corner_index,
                                           points_root = points_root,
                                           cache_array = cache_array)
                    if one_point is not None:
                        cp[kp, jp, ip] = one_point
        return cp

    if cell_kji0 is None:
        cache_cp_array = True
    if hasattr(grid, 'array_corner_points'):
        if cell_kji0 is None:
            return grid.array_corner_points
        return grid.array_corner_points[tuple(cell_kji0)]
    points_root = grid.resolve_geometry_child('Points', child_node = points_root)
    #      if points_root is None: return None  # geometry not present
    if cache_resqml_array:
        grid.point_raw(points_root = points_root, cache_array = True)
    if cache_cp_array:
        grid.array_corner_points = np.zeros((grid.nk, grid.nj, grid.ni, 2, 2, 2, 3))
        points = grid.points_ref()
        if points is None:
            return None  # geometry not present
        if grid.has_split_coordinate_lines:
            grid.create_column_pillar_mapping()
            # todo: replace j,i for loops with numpy broadcasting
            if grid.k_gaps:
                for j in range(grid.nj):
                    for i in range(grid.ni):
                        grid.array_corner_points[:, j, i, 0, 0, 0, :] = points[grid.k_raw_index_array,
                                                                               grid.pillars_for_column[j, i, 0, 0], :]
                        grid.array_corner_points[:, j, i, 1, 0, 0, :] = points[grid.k_raw_index_array + 1,
                                                                               grid.pillars_for_column[j, i, 0, 0], :]
                        grid.array_corner_points[:, j, i, 0, 1, 0, :] = points[grid.k_raw_index_array,
                                                                               grid.pillars_for_column[j, i, 1, 0], :]
                        grid.array_corner_points[:, j, i, 1, 1, 0, :] = points[grid.k_raw_index_array + 1,
                                                                               grid.pillars_for_column[j, i, 1, 0], :]
                        grid.array_corner_points[:, j, i, 0, 0, 1, :] = points[grid.k_raw_index_array,
                                                                               grid.pillars_for_column[j, i, 0, 1], :]
                        grid.array_corner_points[:, j, i, 1, 0, 1, :] = points[grid.k_raw_index_array + 1,
                                                                               grid.pillars_for_column[j, i, 0, 1], :]
                        grid.array_corner_points[:, j, i, 0, 1, 1, :] = points[grid.k_raw_index_array,
                                                                               grid.pillars_for_column[j, i, 1, 1], :]
                        grid.array_corner_points[:, j, i, 1, 1, 1, :] = points[grid.k_raw_index_array + 1,
                                                                               grid.pillars_for_column[j, i, 1, 1], :]
            else:
                for j in range(grid.nj):
                    for i in range(grid.ni):
                        grid.array_corner_points[:, j, i, 0, 0, 0, :] = points[:-1, grid.pillars_for_column[j, i, 0,
                                                                                                            0], :]
                        grid.array_corner_points[:, j, i, 1, 0, 0, :] = points[1:, grid.pillars_for_column[j, i, 0,
                                                                                                           0], :]
                        grid.array_corner_points[:, j, i, 0, 1, 0, :] = points[:-1, grid.pillars_for_column[j, i, 1,
                                                                                                            0], :]
                        grid.array_corner_points[:, j, i, 1, 1, 0, :] = points[1:, grid.pillars_for_column[j, i, 1,
                                                                                                           0], :]
                        grid.array_corner_points[:, j, i, 0, 0, 1, :] = points[:-1, grid.pillars_for_column[j, i, 0,
                                                                                                            1], :]
                        grid.array_corner_points[:, j, i, 1, 0, 1, :] = points[1:, grid.pillars_for_column[j, i, 0,
                                                                                                           1], :]
                        grid.array_corner_points[:, j, i, 0, 1, 1, :] = points[:-1, grid.pillars_for_column[j, i, 1,
                                                                                                            1], :]
                        grid.array_corner_points[:, j, i, 1, 1, 1, :] = points[1:, grid.pillars_for_column[j, i, 1,
                                                                                                           1], :]
        else:
            if grid.k_gaps:
                grid.array_corner_points[:, :, :, 0, 0, 0, :] = points[grid.k_raw_index_array, :-1, :-1, :]
                grid.array_corner_points[:, :, :, 1, 0, 0, :] = points[grid.k_raw_index_array + 1, :-1, :-1, :]
                grid.array_corner_points[:, :, :, 0, 1, 0, :] = points[grid.k_raw_index_array, 1:, :-1, :]
                grid.array_corner_points[:, :, :, 1, 1, 0, :] = points[grid.k_raw_index_array + 1, 1:, :-1, :]
                grid.array_corner_points[:, :, :, 0, 0, 1, :] = points[grid.k_raw_index_array, :-1, 1:, :]
                grid.array_corner_points[:, :, :, 1, 0, 1, :] = points[grid.k_raw_index_array + 1, :-1, 1:, :]
                grid.array_corner_points[:, :, :, 0, 1, 1, :] = points[grid.k_raw_index_array, 1:, 1:, :]
                grid.array_corner_points[:, :, :, 1, 1, 1, :] = points[grid.k_raw_index_array + 1, 1:, 1:, :]
            else:
                grid.array_corner_points[:, :, :, 0, 0, 0, :] = points[:-1, :-1, :-1, :]
                grid.array_corner_points[:, :, :, 1, 0, 0, :] = points[1:, :-1, :-1, :]
                grid.array_corner_points[:, :, :, 0, 1, 0, :] = points[:-1, 1:, :-1, :]
                grid.array_corner_points[:, :, :, 1, 1, 0, :] = points[1:, 1:, :-1, :]
                grid.array_corner_points[:, :, :, 0, 0, 1, :] = points[:-1, :-1, 1:, :]
                grid.array_corner_points[:, :, :, 1, 0, 1, :] = points[1:, :-1, 1:, :]
                grid.array_corner_points[:, :, :, 0, 1, 1, :] = points[:-1, 1:, 1:, :]
                grid.array_corner_points[:, :, :, 1, 1, 1, :] = points[1:, 1:, 1:, :]
    if cell_kji0 is None:
        return grid.array_corner_points
    if not geometry_defined_for_all_cells(grid):
        if not cell_geometry_is_defined(grid, cell_kji0, cache_array = cache_resqml_array):
            return None
    if hasattr(grid, 'array_corner_points'):
        return grid.array_corner_points[tuple(cell_kji0)]
    cp = one_cell_cp(grid, cell_kji0, points_root = points_root, cache_array = cache_resqml_array)
    return cp


def invalidate_corner_points(grid):
    """Deletes cached copy of corner points, if present.

    Use if any pillar geometry changes, or to reclaim memory.
    """
    if hasattr(grid, 'array_corner_points'):
        delattr(grid, 'array_corner_points')


def centre_point(grid, cell_kji0 = None, cache_centre_array = False):
    """Returns centre point of a cell or array of centre points of all cells.

    Optionally cache centre points for all cells.

    arguments:
       cell_kji0 (optional): if present, the (k, j, i) indices of the individual cell for which the
          centre point is required; zero based indexing
       cache_centre_array (boolean, default False): If True, or cell_kji0 is None, an array of centre points
          is generated and added as an attribute of the grid, with attribute name array_centre_point

    returns:
       (x, y, z) 3 element numpy array of floats holding centre point of cell;
       or numpy 3+1D array if cell_kji0 is None

    note:
       resulting coordinates are in the same (local) crs as the grid points

    :meta common:
    """

    if cell_kji0 is None:
        cache_centre_array = True

    # note: this function returns a derived object rather than a native resqml object
    if hasattr(grid, 'array_centre_point'):
        if cell_kji0 is None:
            return grid.array_centre_point
        return grid.array_centre_point[tuple(cell_kji0)]  # could check for nan here and return None
    if cache_centre_array:
        # todo: turn off nan warnings
        grid.array_centre_point = np.empty((grid.nk, grid.nj, grid.ni, 3))
        points = grid.points_ref(masked = False)  # todo: think about masking
        if hasattr(grid, 'array_corner_points'):
            grid.array_centre_point = 0.125 * np.sum(grid.array_corner_points,
                                                     axis = (3, 4, 5))  # mean of eight corner points for each cell
        elif grid.has_split_coordinate_lines:
            # todo: replace j,i for loops with numpy broadcasting
            grid.create_column_pillar_mapping()
            if grid.k_gaps:
                for j in range(grid.nj):
                    for i in range(grid.ni):
                        grid.array_centre_point[:, j, i, :] = 0.125 * (
                            points[grid.k_raw_index_array, grid.pillars_for_column[j, i, 0, 0], :] +
                            points[grid.k_raw_index_array + 1, grid.pillars_for_column[j, i, 0, 0], :] +
                            points[grid.k_raw_index_array, grid.pillars_for_column[j, i, 1, 0], :] +
                            points[grid.k_raw_index_array + 1, grid.pillars_for_column[j, i, 1, 0], :] +
                            points[grid.k_raw_index_array, grid.pillars_for_column[j, i, 0, 1], :] +
                            points[grid.k_raw_index_array + 1, grid.pillars_for_column[j, i, 0, 1], :] +
                            points[grid.k_raw_index_array, grid.pillars_for_column[j, i, 1, 1], :] +
                            points[grid.k_raw_index_array + 1, grid.pillars_for_column[j, i, 1, 1], :])
            else:
                for j in range(grid.nj):
                    for i in range(grid.ni):
                        grid.array_centre_point[:, j,
                                                i, :] = 0.125 * (points[:-1, grid.pillars_for_column[j, i, 0, 0], :] +
                                                                 points[1:, grid.pillars_for_column[j, i, 0, 0], :] +
                                                                 points[:-1, grid.pillars_for_column[j, i, 1, 0], :] +
                                                                 points[1:, grid.pillars_for_column[j, i, 1, 0], :] +
                                                                 points[:-1, grid.pillars_for_column[j, i, 0, 1], :] +
                                                                 points[1:, grid.pillars_for_column[j, i, 0, 1], :] +
                                                                 points[:-1, grid.pillars_for_column[j, i, 1, 1], :] +
                                                                 points[1:, grid.pillars_for_column[j, i, 1, 1], :])
        else:
            if grid.k_gaps:
                grid.array_centre_point[:, :, :, :] = 0.125 * (
                    points[grid.k_raw_index_array, :-1, :-1, :] + points[grid.k_raw_index_array, :-1, 1:, :] +
                    points[grid.k_raw_index_array, 1:, :-1, :] + points[grid.k_raw_index_array, 1:, 1:, :] +
                    points[grid.k_raw_index_array + 1, :-1, :-1, :] + points[grid.k_raw_index_array + 1, :-1, 1:, :] +
                    points[grid.k_raw_index_array + 1, 1:, :-1, :] + points[grid.k_raw_index_array + 1, 1:, 1:, :])
            else:
                grid.array_centre_point[:, :, :, :] = 0.125 * (points[:-1, :-1, :-1, :] + points[:-1, :-1, 1:, :] +
                                                               points[:-1, 1:, :-1, :] + points[:-1, 1:, 1:, :] +
                                                               points[1:, :-1, :-1, :] + points[1:, :-1, 1:, :] +
                                                               points[1:, 1:, :-1, :] + points[1:, 1:, 1:, :])
        if cell_kji0 is None:
            return grid.array_centre_point
        return grid.array_centre_point[cell_kji0[0], cell_kji0[1],
                                       cell_kji0[2]]  # could check for nan here and return None
    cp = grid.corner_points(cell_kji0 = cell_kji0, cache_cp_array = False)
    if cp is None:
        return None
    centre = np.zeros(3)
    for axis in range(3):
        centre[axis] = np.mean(cp[:, :, :, axis])
    return centre


def centre_point_list(grid, cell_kji0s):
    """Returns centre points for a list of cells; caches centre points for all cells.

    arguments:
       cell_kji0s (numpy int array of shape (N, 3)): the (k, j, i) indices of the individual cells for which the
          centre points are required; zero based indexing

    returns:
       numpy float array of shape (N, 3) being the (x, y, z) centre points of the cells

    note:
       resulting coordinates are in the same (local) crs as the grid points
    """

    assert cell_kji0s.ndim == 2 and cell_kji0s.shape[1] == 3
    centres_list = np.empty(cell_kji0s.shape)
    for cell in range(len(cell_kji0s)):
        centres_list[cell] = grid.centre_point(cell_kji0 = cell_kji0s[cell], cache_centre_array = True)
    return centres_list


def point_areally(grid, tolerance = 0.001):
    """Returns array indicating which cells are reduced to a point in both I & J axes.

    Returns:
        numopy bool array of shape extent_kji

    Note:
       Any NaN point values will yield True for a cell
    """

    points = grid.points_ref(masked = False)
    # todo: turn off NaN warning for numpy > ?
    if grid.has_split_coordinate_lines:
        pillar_for_col = grid.create_column_pillar_mapping()
        j_pair_vectors = points[:, pillar_for_col[:, :, 1, :], :] - points[:, pillar_for_col[:, :, 0, :], :]
        i_pair_vectors = points[:, pillar_for_col[:, :, :, 1], :] - points[:, pillar_for_col[:, :, :, 0], :]
        j_pair_nans = np.isnan(j_pair_vectors)
        i_pair_nans = np.isnan(i_pair_vectors)
        any_nans = np.any(np.logical_or(j_pair_nans, i_pair_nans), axis = (3, 4))
        j_pair_extant = np.any(np.abs(j_pair_vectors) > tolerance, axis = -1)
        i_pair_extant = np.any(np.abs(i_pair_vectors) > tolerance, axis = -1)
        any_extant = np.any(np.logical_or(j_pair_extant, i_pair_extant), axis = 3)
    else:
        j_vectors = points[:, 1:, :, :] - points[:, :-1, :, :]
        i_vectors = points[:, :, 1:, :] - points[:, :, :-1, :]
        j_nans = np.any(np.isnan(j_vectors), axis = -1)
        i_nans = np.any(np.isnan(i_vectors), axis = -1)
        j_pair_nans = np.logical_or(j_nans[:, :, :-1], j_nans[:, :, 1:])
        i_pair_nans = np.logical_or(i_nans[:, :-1, :], i_nans[:, 1:, :])
        any_nans = np.logical_or(j_pair_nans, i_pair_nans)
        j_extant = np.any(np.abs(j_vectors) > tolerance, axis = -1)
        i_extant = np.any(np.abs(i_vectors) > tolerance, axis = -1)
        j_pair_extant = np.logical_or(j_extant[:, :, :-1], j_extant[:, :, 1:])
        i_pair_extant = np.logical_or(i_extant[:, :-1, :], i_extant[:, 1:, :])
        any_extant = np.logical_or(j_pair_extant, i_pair_extant)
    layered = np.logical_or(any_nans, np.logical_not(any_extant))
    if grid.k_gaps:
        return np.logical_and(layered[grid.k_raw_index_array], layered[grid.k_raw_index_array + 1])
    return np.logical_and(layered[:-1], layered[1:])


def interpolated_point(grid,
                       cell_kji0,
                       interpolation_fraction,
                       points_root = None,
                       cache_resqml_array = True,
                       cache_cp_array = False):
    """Returns xyz point interpolated from corners of cell.

    Depends on 3 interpolation fractions in range 0 to 1.
    """
    # todo: think about best ordering of axes operations given high aspect ratio of cells (for best accuracy)
    fp = np.empty(3)
    fm = np.empty(3)
    for axis in range(3):
        fp[axis] = max(min(interpolation_fraction[axis], 1.0), 0.0)
        fm[axis] = 1.0 - fp[axis]
    cp = grid.corner_points(cell_kji0,
                            points_root = points_root,
                            cache_resqml_array = cache_resqml_array,
                            cache_cp_array = cache_cp_array)
    c00 = (cp[0, 0, 0] * fm[0] + cp[1, 0, 0] * fp[0])
    c01 = (cp[0, 0, 1] * fm[0] + cp[1, 0, 1] * fp[0])
    c10 = (cp[0, 1, 0] * fm[0] + cp[1, 1, 0] * fp[0])
    c11 = (cp[0, 1, 1] * fm[0] + cp[1, 1, 1] * fp[0])
    c0 = c00 * fm[1] + c10 * fp[1]
    c1 = c01 * fm[1] + c11 * fp[1]
    c = c0 * fm[2] + c1 * fp[2]
    return c


def interpolated_points(grid,
                        cell_kji0,
                        interpolation_fractions,
                        points_root = None,
                        cache_resqml_array = True,
                        cache_cp_array = False):
    """Returns xyz points interpolated from corners of cell.

    Depending on 3 interpolation fraction numpy vectors, each value in range 0 to 1.

    arguments:
       cell_kji0 (triple int): indices of individual cell whose corner points are to be interpolated
       interpolation_fractions (list of three numpy vectors of floats): k, j & i interpolation fraction vectors, each element in range 0 to 1
       points_root (xml node, optional): for efficiency when making multiple calls, this can be set to the xml node of the points data
       cache_resqml_array (boolean, default True): if True, the resqml points data will be cached as an attribute of this grid object
       cache_cp_array (boolean, default False): if True a fully expanded 7D corner points array will be established for this grid and
          cached as an attribute (recommended if looping over many or all the cells and if memory space allows)

    returns:
       4D numpy float array of shape (nik, nij, nii, 3) being the interpolated points; nik is the number of elements in the first of the
       interpolation fraction lists (ie. for k); similarly for nij and nii; the final axis covers xyz

    notea:
       this method returns a lattice of trilinear interpolations of the corner point of the host cell; the returned points are in 'shared'
       arrangement (like resqml points data for an IjkGrid without split pillars or k gaps), not a fully expanded 7D array; calling code
       must redistribute to corner points of individual fine cells if that is the intention
    """

    assert len(interpolation_fractions) == 3
    fp = interpolation_fractions
    fm = []
    for axis in range(3):
        fm.append(1.0 - fp[axis])

    cp = grid.corner_points(cell_kji0,
                            points_root = points_root,
                            cache_resqml_array = cache_resqml_array,
                            cache_cp_array = cache_cp_array)

    c00 = (np.outer(fm[2], cp[0, 0, 0]) + np.outer(fp[2], cp[0, 0, 1]))
    c01 = (np.outer(fm[2], cp[0, 1, 0]) + np.outer(fp[2], cp[0, 1, 1]))
    c10 = (np.outer(fm[2], cp[1, 0, 0]) + np.outer(fp[2], cp[1, 0, 1]))
    c11 = (np.outer(fm[2], cp[1, 1, 0]) + np.outer(fp[2], cp[1, 1, 1]))
    c0 = (np.multiply.outer(fm[1], c00) + np.multiply.outer(fp[1], c01))
    c1 = (np.multiply.outer(fm[1], c10) + np.multiply.outer(fp[1], c11))
    c = (np.multiply.outer(fm[0], c0) + np.multiply.outer(fp[0], c1))

    return c


def find_cell_for_point_xy(grid, x, y, k0 = 0, vertical_ref = 'top', local_coords = True):
    """Searches in 2D for a cell containing point x,y in layer k0; return (j0, i0) or (None, None)."""

    # find minimum of manhatten distances from xy to each corner point
    # then check the four cells around that corner point
    if x is None or y is None:
        return (None, None)
    a = np.array([[x, y, 0.0]])  # extra axis needed to keep global_to_local_crs happy
    if not local_coords:
        grid.global_to_local_crs(a, crs_uuid = grid.crs_uuid)
    a[0, 2] = 0.0  # discard z
    kp = 1 if vertical_ref == 'base' else 0
    (pillar_j0, pillar_i0) = grid.nearest_pillar(a[0, :2], ref_k0 = k0, kp = kp)
    if pillar_j0 > 0 and pillar_i0 > 0:
        cell_kji0 = np.array((k0, pillar_j0 - 1, pillar_i0 - 1))
        poly = grid.poly_line_for_cell(cell_kji0, vertical_ref = vertical_ref)
        if poly is not None and pip.pip_cn(a[0, :2], poly):
            return (cell_kji0[1], cell_kji0[2])
    if pillar_j0 > 0 and pillar_i0 < grid.ni:
        cell_kji0 = np.array((k0, pillar_j0 - 1, pillar_i0))
        poly = grid.poly_line_for_cell(cell_kji0, vertical_ref = vertical_ref)
        if poly is not None and pip.pip_cn(a[0, :2], poly):
            return (cell_kji0[1], cell_kji0[2])
    if pillar_j0 < grid.nj and pillar_i0 > 0:
        cell_kji0 = np.array((k0, pillar_j0, pillar_i0 - 1))
        poly = grid.poly_line_for_cell(cell_kji0, vertical_ref = vertical_ref)
        if poly is not None and pip.pip_cn(a[0, :2], poly):
            return (cell_kji0[1], cell_kji0[2])
    if pillar_j0 < grid.nj and pillar_i0 < grid.ni:
        cell_kji0 = np.array((k0, pillar_j0, pillar_i0))
        poly = grid.poly_line_for_cell(cell_kji0, vertical_ref = vertical_ref)
        if poly is not None and pip.pip_cn(a[0, :2], poly):
            return (cell_kji0[1], cell_kji0[2])
    return (None, None)


def find_cell_for_x_sect_xz(x_sect, x, z):
    """Returns the (k0, j0) or (k0, i0) indices of the cell containing point x,z in the cross section.

    arguments:
       x_sect (numpy float array of shape (nk, nj or ni, 2, 2, 2 or 3): the cross section x,z or x,y,z data
       x (float) x-coordinate of point of interest in the cross section space
       z (float): y-coordinate of  point of interest in the cross section space

    note:
       the x_sect data is in the form returned by x_section_corner_points() or split_gap_x_section_points();
       the 2nd of the returned pair is either a J index or I index, whichever was not the axis specified
       when generating the x_sect data; returns (None, None) if point inclusion not detected; if xyz data is
       provided, the y values are ignored; note that the point of interest x,z coordinates are in the space of
       x_sect, so if rotation has occurred, the x value is no longer an easting and is typically picked off a
       cross section plot
    """

    def test_cell(p, x_sect, k0, ji0):
        poly = np.array([
            x_sect[k0, ji0, 0, 0, 0:3:2], x_sect[k0, ji0, 0, 1, 0:3:2], x_sect[k0, ji0, 1, 1, 0:3:2], x_sect[k0, ji0, 1,
                                                                                                             0, 0:3:2]
        ])
        if np.any(np.isnan(poly)):
            return False
        return pip.pip_cn(p, poly)

    assert x_sect.ndim == 5 and x_sect.shape[2] == 2 and x_sect.shape[3] == 2 and 2 <= x_sect.shape[4] <= 3
    n_k = x_sect.shape[0]
    n_j_or_i = x_sect.shape[1]
    tolerance = 1.0e-3

    if x_sect.shape[4] == 3:
        diffs = x_sect[:, :, :, :, 0:3:2].copy()  # x,z points only
    else:
        diffs = x_sect.copy()
    diffs -= np.array((x, z))
    diffs = np.sum(diffs * diffs, axis = -1)  # square of distance of each point from given x,z
    flat_index = np.nanargmin(diffs)
    min_dist_sqr = diffs.flatten()[flat_index]
    cell_flat_k0_ji0, flat_k_ji_p = divmod(flat_index, 4)
    found_k0, found_ji0 = divmod(cell_flat_k0_ji0, n_j_or_i)
    found_kp, found_jip = divmod(flat_k_ji_p, 2)

    found = test_cell((x, z), x_sect, found_k0, found_ji0)
    if found:
        return found_k0, found_ji0
    # check cells below whilst still close to point
    while found_k0 < n_k - 1:
        found_k0 += 1
        if np.nanmin(diffs[found_k0, found_ji0]) > min_dist_sqr + tolerance:
            break
        found = test_cell((x, z), x_sect, found_k0, found_ji0)
        if found:
            return found_k0, found_ji0

    # try neighbouring column (in case of fault or point exactly on face)
    ji_neighbour = 1 if found_jip == 1 else -1
    found_ji0 += ji_neighbour
    if 0 <= found_ji0 < n_j_or_i:
        col_diffs = diffs[:, found_ji0]
        flat_index = np.nanargmin(col_diffs)
        if col_diffs.flatten()[flat_index] <= min_dist_sqr + tolerance:
            found_k0 = flat_index // 4
            found = test_cell((x, z), x_sect, found_k0, found_ji0)
            if found:
                return found_k0, found_ji0
            # check cells below whilst still close to point
            while found_k0 < n_k - 1:
                found_k0 += 1
                if np.nanmin(diffs[found_k0, found_ji0]) > min_dist_sqr + tolerance:
                    break
                found = test_cell((x, z), x_sect, found_k0, found_ji0)
                if found:
                    return found_k0, found_ji0

    return None, None


def split_horizons_points(grid, min_k0 = None, max_k0 = None, masked = False):
    """Returns reference to a corner points layer of shape (nh, nj, ni, 2, 2, 3) where nh is number of horizons.

    arguments:
       min_k0 (integer): the lowest horizon layer number to be included, in the range 0 to nk + k_gaps; defaults to zero
       max_k0 (integer): the highest horizon layer number to be included, in the range 0 to nk + k_gaps; defaults to nk + k_gaps
       masked (boolean, default False): if True, a masked array is returned with NaN points masked out;
          if False, a simple (unmasked) numpy array is returned

    returns:
       numpy array of shape (nh, nj, ni, 2, 2, 3) where nh = max_k0 - min_k0 + 1, being corner point x,y,z values
       for horizon corners (h, j, i, jp, ip) where h is the horizon (layer interface) index in the range
       0 .. max_k0 - min_k0

    notes:
       data for horizon max_k0 is included in the result (unlike with python ranges);
       in the case of a grid with k gaps, the horizons points returned will follow the k indexing of the points data
       and calling code will need to keep track of the min_k0 offset when using k_raw_index_array to select a slice
       of the horizons points array
    """

    if min_k0 is None:
        min_k0 = 0
    else:
        assert 0 <= min_k0 <= grid.nk_plus_k_gaps
    if max_k0 is None:
        max_k0 = grid.nk_plus_k_gaps
    else:
        assert min_k0 <= max_k0 <= grid.nk_plus_k_gaps
    end_k0 = max_k0 + 1
    points = grid.points_ref(masked = False)
    hp = np.empty((end_k0 - min_k0, grid.nj, grid.ni, 2, 2, 3))
    if grid.has_split_coordinate_lines:
        grid.create_column_pillar_mapping()
        for j in range(grid.nj):
            for i in range(grid.ni):
                hp[:, j, i, 0, 0, :] = points[min_k0:end_k0, grid.pillars_for_column[j, i, 0, 0], :]
                hp[:, j, i, 1, 0, :] = points[min_k0:end_k0, grid.pillars_for_column[j, i, 1, 0], :]
                hp[:, j, i, 0, 1, :] = points[min_k0:end_k0, grid.pillars_for_column[j, i, 0, 1], :]
                hp[:, j, i, 1, 1, :] = points[min_k0:end_k0, grid.pillars_for_column[j, i, 1, 1], :]
    else:
        hp[:, :, :, 0, 0, :] = points[min_k0:end_k0, :-1, :-1, :]
        hp[:, :, :, 1, 0, :] = points[min_k0:end_k0, 1:, :-1, :]
        hp[:, :, :, 0, 1, :] = points[min_k0:end_k0, :-1, 1:, :]
        hp[:, :, :, 1, 1, :] = points[min_k0:end_k0, 1:, 1:, :]
    return hp



"""A submodule containing functions relating to xyz grid coordinates"""

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.olio.xml_et as rqet
import resqpy.weights_and_measures as bwam
import resqpy.crs as rqc
import resqpy.olio.vector_utilities as vec


def xyz_box(grid, points_root = None, lazy = True, local = False):
    """Returns the minimum and maximum xyz for the grid geometry.

    arguments:
       points_root (optional): if not None, the xml root node for the points data (speed optimization)
       lazy (boolean, default True): if True, only the 8 outermost logical corners of the grid are used
          to determine the ranges of xyz; if False, all the points in the entire grid are scanned to
          determine the xyz ranges in an exhaustive manner
       local (boolean, default False): if True, the xyz ranges that are returned are in the local
          coordinate space, otherwise the global (crs parent) coordinate space

    returns:
       numpy array of float of shape (2, 3); the first axis is minimum, maximum; the second axis is x, y, z

    note:
       if the lazy argument is True, the results are likely to under-report the ranges, especially for z

    :meta common:
    """

    if grid.xyz_box_cached is None or (not lazy and not grid.xyz_box_cached_thoroughly):
        grid.xyz_box_cached = np.zeros((2, 3))
        if lazy:
            eight_corners = np.zeros((2, 2, 2, 3))
            for kp in [0, 1]:
                for jp in [0, 1]:
                    for ip in [0, 1]:
                        cell_kji0 = [
                            kp * (grid.extent_kji[0] - 1), jp * (grid.extent_kji[1] - 1), ip * (grid.extent_kji[2] - 1)
                        ]
                        eight_corners[kp, jp, ip] = grid.point(cell_kji0 = cell_kji0,
                                                               corner_index = [kp, jp, ip],
                                                               points_root = points_root,
                                                               cache_array = False)
            grid.xyz_box_cached[0, :] = np.nanmin(eight_corners, axis = (0, 1, 2))
            grid.xyz_box_cached[1, :] = np.nanmax(eight_corners, axis = (0, 1, 2))
        else:
            ps = grid.points_ref()
            if grid.has_split_coordinate_lines:
                grid.xyz_box_cached[0, :] = np.nanmin(ps, axis = (0, 1))
                grid.xyz_box_cached[1, :] = np.nanmax(ps, axis = (0, 1))
            else:
                grid.xyz_box_cached[0, :] = np.nanmin(ps, axis = (0, 1, 2))
                grid.xyz_box_cached[1, :] = np.nanmax(ps, axis = (0, 1, 2))
        grid.xyz_box_cached_thoroughly = not lazy
    if local:
        return grid.xyz_box_cached
    global_xyz_box = grid.xyz_box_cached.copy()
    return grid.local_to_global_crs(global_xyz_box, crs_uuid = grid.crs_uuid)


def xyz_box_centre(grid, points_root = None, lazy = False, local = False):
    """Returns the (x,y,z) point (as 3 element numpy) at the centre of the xyz box for the grid.

    arguments:
       points_root (optional): if not None, the xml root node for the points data (speed optimization)
       lazy (boolean, default True): if True, only the 8 outermost logical corners of the grid are used
          to determine the ranges of xyz and hence the centre; if False, all the points in the entire
          grid are scanned to determine the xyz ranges in an exhaustive manner
       local (boolean, default False): if True, the xyz values that are returned are in the local
          coordinate space, otherwise the global (crs parent) coordinate space

    returns:
       numpy array of float of shape (3,) being the x, y, z coordinates of the centre of the grid

    note:
       the centre point returned is simply the midpoint of the x, y & z ranges of the grid
    """

    return np.nanmean(grid.xyz_box(points_root = points_root, lazy = lazy, local = local), axis = 0)


def bounding_box(grid, cell_kji0, points_root = None, cache_cp_array = False):
    """Returns the xyz box which envelopes the specified cell, as a numpy array of shape (2, 3)."""

    result = np.zeros((2, 3))
    cp = grid.corner_points(cell_kji0, points_root = points_root, cache_cp_array = cache_cp_array)
    result[0] = np.min(cp, axis = (0, 1, 2))
    result[1] = np.max(cp, axis = (0, 1, 2))
    return result


def composite_bounding_box(grid, bounding_box_list):
    """Returns the xyz box which envelopes all the boxes in the list, as a numpy array of shape (2, 3)."""

    result = bounding_box_list[0]
    for box in bounding_box_list[1:]:
        result[0] = np.minimum(result[0], box[0])
        result[1] = np.maximum(result[1], box[1])
    return result


def _local_to_global_crs(grid,
                         a,
                         crs_uuid = None,
                         global_xy_units = None,
                         global_z_units = None,
                         global_z_increasing_downward = None):
    """Converts array of points in situ from local coordinate system to global one."""

    if crs_uuid is None:
        crs_uuid = grid.crs_uuid
        if crs_uuid is None:
            return a

    crs = rqc.Crs(grid.model, uuid = crs_uuid)

    if crs.rotated:
        a[:] = vec.rotate_array(crs.rotation_matrix, a)

    flat_a = a.reshape((-1, 3))  # flattened view of array a as vector of (x, y, z) points, in situ

    # note: here negation is made in local crs; if z_offset is not zero, this might not be what is intended
    if global_z_increasing_downward is not None:
        if global_z_increasing_downward != crs.z_inc_down:
            flat_a[:, 2] = np.negative(flat_a[:, 2])

    # This code assumes x, y, z offsets are in local crs units
    flat_a[:, 0] += crs.x_offset
    flat_a[:, 1] += crs.y_offset
    flat_a[:, 2] += crs.z_offset

    if global_xy_units is not None:
        bwam.convert_lengths(flat_a[:, 0], crs.xy_units, global_xy_units)  # x
        bwam.convert_lengths(flat_a[:, 1], crs.xy_units, global_xy_units)  # y
    if global_z_units is not None:
        bwam.convert_lengths(flat_a[:, 2], crs.z_units, global_z_units)  # z

    return flat_a.reshape(a.shape)


def z_inc_down(grid):
    """Return True if z increases downwards in the coordinate reference system used by the grid geometry

    :meta common:
    """

    if grid.crs is None:
        assert grid.crs_uuid is not None
        grid.crs = rqc.Crs(grid.model, uuid = grid.crs_uuid)
    return grid.crs.z_inc_down


def _global_to_local_crs(grid,
                         a,
                         crs_uuid = None,
                         global_xy_units = None,
                         global_z_units = None,
                         global_z_increasing_downward = None):
    """Converts array of points in situ from global coordinate system to established local one."""

    if crs_uuid is None:
        crs_uuid = grid.crs_uuid
        if crs_uuid is None:
            return a

    flat_a = a.reshape((-1, 3))  # flattened view of array a as vector of (x, y, z) points, in situ

    crs = rqc.Crs(grid.model, uuid = crs_uuid)

    if global_xy_units is not None:
        bwam.convert_lengths(flat_a[:, 0], global_xy_units, crs.xy_units)  # x
        bwam.convert_lengths(flat_a[:, 1], global_xy_units, crs.xy_units)  # y
    if global_z_units is not None:
        bwam.convert_lengths(flat_a[:, 2], global_z_units, crs.z_units)  # z

    # This code assumes x, y, z offsets are in local crs units
    flat_a[:, 0] -= crs.x_offset
    flat_a[:, 1] -= crs.y_offset
    flat_a[:, 2] -= crs.z_offset

    # note: here negation is made in local crs; if z_offset is not zero, this might not be what is intended
    if global_z_increasing_downward is not None:
        if global_z_increasing_downward != crs.z_inc_down:
            flat_a[:, 2] = np.negative(flat_a[:, 2])

    if crs.rotated:
        flat_a[:] = vec.rotate_array(crs.reverse_rotation_matrix, flat_a)

    return flat_a.reshape(a.shape)


def check_top_and_base_cell_edge_directions(grid):
    """Check grid top face I & J edge vectors (in x,y) against basal equivalents.

    Max 90 degree angle tolerated.

    returns:
        boolean: True if all checks pass; False if one or more checks fail

    notes:
       similarly checks cell edge directions in neighbouring cells in top (and separately in base)
       currently requires geometry to be defined for all pillars
       logs a warning if a check is not passed
    """

    log.debug('deriving cell edge vectors at top and base (for checking)')
    grid.point(cache_array = True)
    good = True
    if grid.has_split_coordinate_lines:
        # build top and base I & J cell edge vectors
        grid.create_column_pillar_mapping()  # pillar indices for 4 columns around interior pillars
        top_j_edge_vectors_p = np.zeros((grid.nj, grid.ni, 2, 2))  # third axis is ip
        top_i_edge_vectors_p = np.zeros((grid.nj, 2, grid.ni, 2))  # second axis is jp
        base_j_edge_vectors_p = np.zeros((grid.nj, grid.ni, 2, 2))  # third axis is ip
        base_i_edge_vectors_p = np.zeros((grid.nj, 2, grid.ni, 2))  # second axis is jp
        # todo: rework as numpy operations across nj & ni
        for j in range(grid.nj):
            for i in range(grid.ni):
                for jip in range(2):  # serves as either jp or ip
                    top_j_edge_vectors_p[j, i,
                                         jip, :] = (grid.points_cached[0, grid.pillars_for_column[j, i, 1, jip], :2] -
                                                    grid.points_cached[0, grid.pillars_for_column[j, i, 0, jip], :2])
                    base_j_edge_vectors_p[j, i, jip, :] = (
                        grid.points_cached[grid.nk_plus_k_gaps, grid.pillars_for_column[j, i, 1, jip], :2] -
                        grid.points_cached[grid.nk_plus_k_gaps, grid.pillars_for_column[j, i, 0, jip], :2])
                    top_i_edge_vectors_p[j, jip,
                                         i, :] = (grid.points_cached[0, grid.pillars_for_column[j, i, jip, 1], :2] -
                                                  grid.points_cached[0, grid.pillars_for_column[j, i, jip, 0], :2])
                    base_i_edge_vectors_p[j, jip, i, :] = (
                        grid.points_cached[grid.nk_plus_k_gaps, grid.pillars_for_column[j, i, jip, 1], :2] -
                        grid.points_cached[grid.nk_plus_k_gaps, grid.pillars_for_column[j, i, jip, 0], :2])
        # reshape to allow common checking code with unsplit grid vectors (below)
        top_j_edge_vectors = top_j_edge_vectors_p.reshape((grid.nj, 2 * grid.ni, 2))
        top_i_edge_vectors = top_i_edge_vectors_p.reshape((2 * grid.nj, grid.ni, 2))
        base_j_edge_vectors = base_j_edge_vectors_p.reshape((grid.nj, 2 * grid.ni, 2))
        base_i_edge_vectors = base_i_edge_vectors_p.reshape((2 * grid.nj, grid.ni, 2))
    else:
        top_j_edge_vectors = (grid.points_cached[0, 1:, :, :2] - grid.points_cached[0, :-1, :, :2]).reshape(
            (grid.nj, grid.ni + 1, 2))
        top_i_edge_vectors = (grid.points_cached[0, :, 1:, :2] - grid.points_cached[0, :, :-1, :2]).reshape(
            (grid.nj + 1, grid.ni, 2))
        base_j_edge_vectors = (grid.points_cached[-1, 1:, :, :2] - grid.points_cached[-1, :-1, :, :2]).reshape(
            (grid.nj, grid.ni + 1, 2))
        base_i_edge_vectors = (grid.points_cached[-1, :, 1:, :2] - grid.points_cached[-1, :, :-1, :2]).reshape(
            (grid.nj + 1, grid.ni, 2))
    log.debug('checking relative direction of top and base edges')
    # check direction of top edges against corresponding base edges, tolerate upto 90 degree difference
    dot_j = np.sum(top_j_edge_vectors * base_j_edge_vectors, axis = 2)
    dot_i = np.sum(top_i_edge_vectors * base_i_edge_vectors, axis = 2)
    if not np.all(dot_j >= 0.0) and np.all(dot_i >= 0.0):
        log.warning('one or more columns of cell edges flip direction: this grid is probably unusable')
        good = False
    log.debug('checking relative direction of edges in neighbouring cells at top of grid (and base)')
    # check direction of similar edges on neighbouring cells, tolerate upto 90 degree difference
    dot_jp = np.sum(top_j_edge_vectors[1:, :, :] * top_j_edge_vectors[:-1, :, :], axis = 2)
    dot_ip = np.sum(top_i_edge_vectors[1:, :, :] * top_i_edge_vectors[:-1, :, :], axis = 2)
    if not np.all(dot_jp >= 0.0) and np.all(dot_ip >= 0.0):
        log.warning('top cell edges for neighbouring cells flip direction somewhere: this grid is probably unusable')
        good = False
    dot_jp = np.sum(base_j_edge_vectors[1:, :, :] * base_j_edge_vectors[:-1, :, :], axis = 2)
    dot_ip = np.sum(base_i_edge_vectors[1:, :, :] * base_i_edge_vectors[:-1, :, :], axis = 2)
    if not np.all(dot_jp >= 0.0) and np.all(dot_ip >= 0.0):
        log.warning('base cell edges for neighbouring cells flip direction somewhere: this grid is probably unusable')
        good = False
    return good



"""_import_nexus.py: Module to import a nexus corp grid & properties, or vdb, or vdb ensemble into resqml format."""

version = '15th November 2021'

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import glob
import os

import numpy as np

import resqpy.crs as rqc
import resqpy.model as rq
import resqpy.olio.ab_toolbox as abt
import resqpy.olio.load_data as ld
import resqpy.olio.trademark as tm
import resqpy.olio.vdb as vdb
import resqpy.olio.write_data as wd
import resqpy.olio.xml_et as rqet
import resqpy.property as rp
import resqpy.time_series as rts
import resqpy.weights_and_measures as bwam

from resqpy.rq_import import grid_from_cp


def import_nexus(
        resqml_file_root,  # output path and file name without .epc or .h5 extension
        extent_ijk = None,  # 3 element numpy vector
        vdb_file = None,  # vdb input file: either this or corp_file should be not None
        vdb_case = None,  # if None, first case in vdb is used (usually a vdb only holds one case)
        corp_file = None,  # corp ascii input file: nexus corp data without keyword
        corp_bin_file = None,  # corp binary file: nexus corp data in bespoke binary format
        corp_xy_units = 'm',
        corp_z_units = 'm',
        corp_z_inc_down = True,
        ijk_handedness = 'right',
        corp_eight_mode = False,
        geometry_defined_everywhere = True,
        treat_as_nan = None,
        active_mask_file = None,
        use_binary = False,  # this refers to pure binary arrays, not corp bin format
        resqml_xy_units = 'm',
        resqml_z_units = 'm',
        resqml_z_inc_down = True,
        shift_to_local = False,
        local_origin_place = 'centre',  # 'centre' or 'minimum'
        max_z_void = 0.1,  # vertical gaps greater than this will introduce k gaps intp resqml grid
        split_pillars = True,
        split_tolerance = 0.01,  # applies to each of x, y, z differences
        property_array_files = None,  # actually, list of (filename, keyword, uom, time_index, null_value, discrete)
        summary_file = None,  # used to extract timestep dates when loading recurrent data from vdb
        vdb_static_properties = True,
        # if True, static vdb properties are imported (only relevant if vdb_file is not None)
        vdb_recurrent_properties = False,
        timestep_selection = 'all',
        # 'first', 'last', 'first and last', 'all', or list of ints being reporting timestep numbers
        use_compressed_time_series = True,
        decoarsen = True,  # where ICOARSE is present, redistribute data to uncoarse cells
        ab_property_list = None,
        # list of (file_name, keyword, property_kind, facet_type, facet, uom, time_index, null_value, discrete)
        create_property_set = False,
        ensemble_case_dirs_root = None,  # path upto but excluding realisation number
        ensemble_property_dictionary = None,
        # dictionary mapping title (or keyword) to (filename, property_kind, facet_type, facet,
        #                                           uom, time_index, null_value, discrete)
        ensemble_size_limit = None,
        grid_title = 'ROOT',
        mode = 'w',
        progress_fn = None):
    """Read a simulation grid geometry and optionally grid properties.

    Input may be from nexus ascii input files, or nexus vdb output.

    Arguments:
        resqml_file_root (str): output path and file name without .epc or .h5 extension
        extent_ijk (triple float, optional): ijk extents (fortran ordering)
        vdb_file (str, optional): vdb input file, either this or corp_file should be not None. Required if importing from a vdb
        vdb_case (str, optional): required if the vdb contains more than one case. If None, first case in vdb is used
        corp_file (str, optional): required if importing from corp ascii file. corp ascii input file: nexus corp data without keyword
        corp_bin_file (str, optional): required if importing from corp binary file
        corp_xy_units (str, default 'm'): xy length units
        corp_z_units (str, default 'm'): z length units
        corp_z_inc_down (bool, default True): if True z values increase with depth
        ijk_handedness (str, default 'right'): 'right' or 'left'
        corp_eight_mode (bool, default False): if True the ordering of corner point data is in nexus EIGHT mode
        geometry_defined_everywhere (bool, default True): if False then inactive cells are marked as not having geometry
        treat_as_nan (float, default None): if a value is provided corner points with this value will be assigned nan
        active_mask_file (str, default None): ascii property file holding values 0 or 1, with 1 indicating active cells
        use_binary (bool, default False): if True a cached binary version of ascii files will be used (pure binary, not corp bin format)
        resqml_xy_units (str, default 'm'): output xy units for resqml file
        resqml_z_units (str, default 'm'): output z units for resqml file
        resqml_z_inc_down (bool, default True): if True z values increase with depth for output resqml file
        shift_to_local (bool, default False): if True then a local origin will be used in the CRS
        local_origin_place (str, default 'centre'): 'centre' or 'minimum'. If 'centre' the local origin is placed at the centre of the grid; ignored if shift_to_local is False
        max_z_void (float, default 0.1): maximum z gap between vertically neighbouring corner points. Vertical gaps greater than this will introduce k gaps into resqml grid. Units are corp z units
        split_pillars (bool, default True): if False an unfaulted grid will be generated
        split_tolerance (float, default 0.01): maximum distance between neighbouring corner points before a pillar is considered 'split'. Applies to each of x, y, z differences
        property_array_files (list, default None): list of (filename, keyword, uom, time_index, null_value, discrete)
        summary_file (str, default None): nexus output summary file, used to extract timestep dates when loading recurrent data from vdb
        vdb_static_properties (bool, default True): if True, static vdb properties are imported (only relevant if vdb_file is not None)
        vdb_recurrent_properties (bool, default False): # if True, recurrent vdb properties are imported (only relevant if vdb_file is not None)
        timestep_selection (str, default 'all): 'first', 'last', 'first and last', 'all', or list of ints being reporting timestep numbers. Ignored if vdb_recurrent_properties is False
        use_compressed_time_series (bool, default True): generates reduced time series containing timesteps with recurrent properties from vdb, rather than full nexus summary time series
        decoarsen (bool, default True): where ICOARSE is present, redistribute data to uncoarse cells
        ab_property_list (list, default None):  list of (file_name, keyword, property_kind, facet_type, facet, uom, time_index, null_value, discrete)
        create_property_set (bool, default False): if True a resqml PropertySet is created
        ensemble_case_dirs_root (str, default None): path up to but excluding realisation number
        ensemble_property_dictionary (str, default None): dictionary mapping title (or keyword) to (filename, property_kind, facet_type, facet, uom, time_index, null_value, discrete)
        ensemble_size_limit (int, default None): if present processing of ensemble will terminate after this number of cases is reached
        grid_title (str, default 'ROOT'): grid citation title
        mode (str, default 'w'): 'w' or 'a', mode to write or append to hdf5
        progress_fn (function, default None): if present function must have one floating argument with value increasing from 0 to 1, and is called at intervals to indicate progress

    Returns:
        resqml model in memory & written to disc
    """

    if resqml_file_root.endswith('.epc'):
        resqml_file_root = resqml_file_root[:-4]
    assert mode in ['w', 'a']

    if vdb_file:
        using_vdb = True
        corp_file = corp_bin_file = None
        grid_title = grid_title.upper()
        log.info('starting import of Nexus ' + str(grid_title) + ' corp from vdb ' + str(vdb_file))
        tm.log_nexus_tm('info')
        vdbase = vdb.VDB(vdb_file)
        case_list = vdbase.cases()
        assert len(case_list) > 0, 'no cases found in vdb'
        if vdb_case is None:
            vdb_case = case_list[0]
        else:
            assert vdb_case in case_list, 'case ' + vdb_case + ' not found in vdb: ' + vdb_file
            vdbase.set_use_case(vdb_case)
        assert grid_title in vdbase.list_of_grids(), 'grid ' + str(grid_title) + ' not found in vdb'
        if extent_ijk is not None:
            vdbase.set_extent_kji(tuple(reversed(extent_ijk)))
        log.debug('using case ' + vdb_case + ' and grid ' + grid_title + ' from vdb')
        if vdb_recurrent_properties and not summary_file:
            if vdb_file.endswith('.vdb.zip'):
                summary_file = vdb_file[:-8] + '.sum'
            elif vdb_file.endswith('.vdb') or vdb_file.endswith('.zip'):
                summary_file = vdb_file[:-4] + '.sum'
            else:
                sep = vdb_file.rfind(os.sep)
                dot = vdb_file[sep + 1:].find('.')
                if dot > 0:
                    summary_file = vdb_file[:sep + 1 + dot] + ',sum'
                else:
                    summary_file = vdb_file + '.sum'
        cp_array = vdbase.grid_corp(grid_title)
        cp_extent_kji = cp_array.shape[:3]
        if cp_extent_kji[:2] == (1, 1):  # auto determination of extent failed
            assert extent_ijk is not None, 'failed to determine extent of grid from corp data'
            (ni, nj, nk) = extent_ijk
            assert cp_extent_kji[2] == ni * nj * nk, 'number of cells in grid corp does not match extent'
            cp_extent = (nk, nj, ni, 2, 2, 2, 3)  # (nk, nj, ni, kp, jp, ip, xyz)
            cp_array = cp_array.reshape(cp_extent)
        elif extent_ijk is not None:
            for axis in range(3):
                assert cp_extent_kji[axis] == extent_ijk[
                    2 - axis], 'extent of grid corp data from vdb does not match that supplied'

    elif corp_file or corp_bin_file:
        if corp_bin_file:
            corp_file = None
        using_vdb = False
        #     geometry_defined_everywhere = (active_mask_file is None)
        log.info('starting import of Nexus corp file ' + str(corp_file if corp_file else corp_bin_file))
        tm.log_nexus_tm('info')
        if extent_ijk is None:  # auto detect extent
            extent_kji = None
            cp_extent = None
        else:
            (ni, nj, nk) = extent_ijk
            extent_kji = np.array((nk, nj, ni), dtype = 'int')
            cp_extent = (nk, nj, ni, 2, 2, 2, 3)  # (nk, nj, ni, kp, jp, ip, xyz)
        log.debug('reading and resequencing corp data')
        if corp_bin_file:  # bespoke nexus corp bin format, not to be confused with pure binary files used below
            cp_array = ld.load_corp_array_from_file(
                corp_bin_file,
                extent_kji,
                corp_bin = True,
                comment_char = None,  # comment char will be detected automatically
                data_free_of_comments = False,
                use_binary = use_binary)
        else:
            cp_binary_file = abt.cp_binary_filename(
                corp_file, nexus_ordering = False)  # pure binary, not bespoke corp bin used above
            recent_binary_exists = ld.file_exists(cp_binary_file, must_be_more_recent_than_file = corp_file)
            cp_array = None
            if use_binary and (extent_ijk is not None) and recent_binary_exists:
                try:
                    cp_array = ld.load_array_from_file(cp_binary_file, cp_extent, use_binary = True)
                except Exception:
                    cp_array = None
            if cp_array is None:
                cp_array = ld.load_corp_array_from_file(
                    corp_file,
                    extent_kji,
                    corp_bin = False,
                    comment_char = None,  # comment char will be detected automatically
                    data_free_of_comments = False,
                    use_binary = use_binary)
                if use_binary:
                    wd.write_pure_binary_data(cp_binary_file,
                                              cp_array)  # NB: this binary file is resequenced, not in nexus ordering!

    else:
        raise ValueError('vdb_file and corp_file are both None in import_nexus() call')

    if cp_array is None:
        log.error('failed to create corner point array')
        return None

    if extent_ijk is None:
        cp_extent = cp_array.shape
        extent_kji = cp_extent[:3]
        (nk, nj, ni) = extent_kji
        extent_ijk = (ni, nj, nk)
    else:
        ni, nj, nk = extent_ijk

    # convert units
    log.debug('Converting units')
    if corp_xy_units == corp_z_units and resqml_xy_units == resqml_z_units:
        bwam.convert_lengths(cp_array, corp_xy_units, resqml_xy_units)
    else:
        bwam.convert_lengths(cp_array[:, :, :, :, :, :, 0:1], corp_xy_units, resqml_xy_units)
        bwam.convert_lengths(cp_array[:, :, :, :, :, :, 2], corp_z_units, resqml_z_units)

    # invert z if required
    if resqml_z_inc_down != corp_z_inc_down:
        log.debug('Inverting z values')
        inversion = np.negative(cp_array[:, :, :, :, :, :, 2])
        cp_array[:, :, :, :, :, :, 2] = inversion

    # read active cell mask
    log.debug('Setting up active cell mask')
    active_mask = inactive_mask = None
    if vdb_file:
        assert vdbase is not None, 'problem with vdb object'
        inactive_mask = vdbase.grid_kid_inactive_mask(grid_title)  # TODO: check conversion of KID to boolean for LGRs
        if inactive_mask is not None:
            log.debug('using kid array as inactive cell mask')
            active_mask = np.logical_not(inactive_mask)
        else:
            log.warning('kid array not found, using unpack array as active cell indicator')
            unp = vdbase.grid_unpack(grid_title)
            assert unp is not None, 'failed to load active cell indicator mask from vdb kid or unpack arrays'
            active_mask = np.empty((nk, nj, ni), dtype = 'bool')
            active_mask[:] = (unp > 0)
            inactive_mask = np.logical_not(active_mask)
    elif active_mask_file:
        active_mask = ld.load_array_from_file(active_mask_file, extent_kji, data_type = 'bool', use_binary = use_binary)
        if active_mask is None:
            log.error('failed to load active cell indicator array from file: ' + active_mask_file)
        else:
            inactive_mask = np.logical_not(active_mask)  # will crash if active mask load failed

    # shift grid geometry to local crs
    local_origin = np.zeros(3)
    if shift_to_local:
        log.debug('shifting to local origin at ' + local_origin_place)
        if local_origin_place == 'centre':
            local_origin = np.nanmean(cp_array, axis = (0, 1, 2, 3, 4, 5))
        elif local_origin_place == 'minimum':
            local_origin = np.nanmin(cp_array, axis = (0, 1, 2, 3, 4, 5)) - 1.0  # The -1 ensures all coords are >0
        else:
            assert (False)
        cp_array -= local_origin

    # create empty resqml model
    log.debug('creating an empty resqml model')
    if mode == 'w':
        model = rq.Model(resqml_file_root, new_epc = True, create_basics = True, create_hdf5_ext = True)
    else:
        model = rq.Model(resqml_file_root)
    assert model is not None
    ext_uuid = model.h5_uuid()
    assert ext_uuid is not None

    # create coodinate reference system (crs) in model and set references in grid object
    log.debug('creating coordinate reference system')
    crs_uuids = model.uuids(obj_type = 'LocalDepth3dCrs')
    new_crs = rqc.Crs(model,
                      x_offset = local_origin[0],
                      y_offset = local_origin[1],
                      z_offset = local_origin[2],
                      xy_units = resqml_xy_units,
                      z_units = resqml_z_units,
                      z_inc_down = resqml_z_inc_down)
    new_crs.create_xml(reuse = True)
    crs_uuid = new_crs.uuid

    grid = grid_from_cp(model,
                        cp_array,
                        crs_uuid,
                        active_mask = active_mask,
                        geometry_defined_everywhere = geometry_defined_everywhere,
                        treat_as_nan = treat_as_nan,
                        max_z_void = max_z_void,
                        split_pillars = split_pillars,
                        split_tolerance = split_tolerance,
                        ijk_handedness = ijk_handedness,
                        known_to_be_straight = False)

    # create hdf5 file using arrays cached in grid above
    log.info('writing grid geometry to hdf5 file ' + resqml_file_root + '.h5')
    grid.write_hdf5_from_caches(resqml_file_root + '.h5', mode = mode, write_active = False)

    # build xml for grid geometry
    log.debug('building xml for grid')
    ijk_node = grid.create_xml(ext_uuid = None, title = grid_title, add_as_part = True, add_relationships = True)
    assert ijk_node is not None, 'failed to create IjkGrid node in xml tree'

    # impprt property arrays into a collection
    prop_import_collection = None
    decoarsen_array = None
    ts_node = None
    ts_uuid = None

    if active_mask is None and grid.inactive is not None:
        active_mask = np.logical_not(grid.inactive)

    if using_vdb:
        prop_import_collection = rp.GridPropertyCollection()
        if vdb_static_properties:
            props = vdbase.grid_list_of_static_properties(grid_title)
            if len(props) > 0:
                prop_import_collection = rp.GridPropertyCollection()
                prop_import_collection.set_grid(grid)
                for keyword in props:
                    prop_import_collection.import_vdb_static_property_to_cache(vdbase, keyword, grid_name = grid_title)
    #      if active_mask is not None:
    #         prop_import_collection.add_cached_array_to_imported_list(active_mask, active_mask_file, 'ACTIVE', property_kind = 'active',
    #                                                                  discrete = True, uom = None, time_index = None, null_value = None)

    elif property_array_files is not None and len(property_array_files) > 0:
        prop_import_collection = rp.GridPropertyCollection()
        prop_import_collection.set_grid(grid)
        for (p_filename, p_keyword, p_uom, p_time_index, p_null_value, p_discrete) in property_array_files:
            prop_import_collection.import_nexus_property_to_cache(p_filename,
                                                                  p_keyword,
                                                                  grid.extent_kji,
                                                                  discrete = p_discrete,
                                                                  uom = p_uom,
                                                                  time_index = p_time_index,
                                                                  null_value = p_null_value,
                                                                  use_binary = use_binary)
    #      if active_mask is not None:
    #         prop_import_collection.add_cached_array_to_imported_list(active_mask, active_mask_file, 'ACTIVE', property_kind = 'active',
    #                                                                  discrete = True, uom = None, time_index = None, null_value = None)

    #  ab_property_list: list of (filename, keyword, property_kind, facet_type, facet, uom, time_index, null_value, discrete)
    elif ab_property_list is not None and len(ab_property_list) > 0:
        prop_import_collection = rp.GridPropertyCollection()
        prop_import_collection.set_grid(grid)
        for (p_filename, p_keyword, p_property_kind, p_facet_type, p_facet, p_uom, p_time_index, p_null_value,
             p_discrete) in ab_property_list:
            prop_import_collection.import_ab_property_to_cache(p_filename,
                                                               p_keyword,
                                                               grid.extent_kji,
                                                               discrete = p_discrete,
                                                               property_kind = p_property_kind,
                                                               facet_type = p_facet_type,
                                                               facet = p_facet,
                                                               uom = p_uom,
                                                               time_index = p_time_index,
                                                               null_value = p_null_value)
    #      if active_mask is not None:
    #         prop_import_collection.add_cached_array_to_imported_list(active_mask, active_mask_file, 'ACTIVE', property_kind = 'active',
    #                                                                  discrete = True, uom = None, time_index = None, null_value = None)

    # ensemble_property_dictionary: mapping title (or keyword) to
    #    (filename, property_kind, facet_type, facet, uom, time_index, null_value, discrete)
    elif ensemble_case_dirs_root and ensemble_property_dictionary:
        case_path_list = glob.glob(ensemble_case_dirs_root + '*')
        assert len(case_path_list) > 0, 'no case directories found with path starting: ' + str(ensemble_case_dirs_root)
        case_number_place = len(ensemble_case_dirs_root)
        case_zero_used = False
        case_count = 0
        for case_path in case_path_list:
            if ensemble_size_limit is not None and case_count >= ensemble_size_limit:
                log.warning('stopping after reaching ensemble size limit')
                break
            # NB. import each case individually rather than holding property arrays for whole ensemble in memory at once
            prop_import_collection = rp.GridPropertyCollection()
            prop_import_collection.set_grid(grid)
            tail = case_path[case_number_place:]
            try:
                case_number = int(tail)
                assert case_number >= 0, 'negative case number encountered'
                if case_number == 0:
                    assert not case_zero_used, 'more than one case number evaluated to zero'
                    case_zero_used = True
            except Exception:
                log.error('failed to determine case number for tail: ' + str(tail))
                continue
            for keyword in ensemble_property_dictionary.keys():
                (filename, p_property_kind, p_facet_type, p_facet, p_uom, p_time_index, p_null_value,
                 p_discrete) = ensemble_property_dictionary[keyword]
                p_filename = os.path.join(case_path, filename)
                if not os.path.exists(p_filename):
                    log.error('missing property file: ' + p_filename)
                    continue
                prop_import_collection.import_nexus_property_to_cache(p_filename,
                                                                      keyword,
                                                                      grid.extent_kji,
                                                                      discrete = p_discrete,
                                                                      uom = p_uom,
                                                                      time_index = p_time_index,
                                                                      null_value = p_null_value,
                                                                      property_kind = p_property_kind,
                                                                      facet_type = p_facet_type,
                                                                      facet = p_facet,
                                                                      realization = case_number,
                                                                      use_binary = False)
            if len(prop_import_collection.imported_list) > 0:
                # create hdf5 file using arrays cached in grid above
                log.info('writing properties to hdf5 file ' + str(resqml_file_root) + '.h5 for case: ' +
                         str(case_number))
                grid.write_hdf5_from_caches(resqml_file_root + '.h5',
                                            geometry = False,
                                            imported_properties = prop_import_collection,
                                            write_active = False)
                # add imported properties parts to model, building property parts list
                prop_import_collection.create_xml_for_imported_list_and_add_parts_to_model(ext_uuid,
                                                                                           time_series_uuid = ts_uuid)
                if create_property_set:
                    prop_import_collection.create_property_set_xml('realisation ' + str(case_number))
                case_count += 1
            # remove cached static property arrays from memory

            #         prop_import_collection.remove_all_cached_arrays()
            del prop_import_collection
            prop_import_collection = None
        log.info(f'Nexus ascii ensemble input processed {case_count} cases')
        tm.log_nexus_tm('info')

    # create hdf5 file using arrays cached in grid above
    if prop_import_collection is not None and len(prop_import_collection.imported_list) > 0:
        if decoarsen:
            decoarsen_array = prop_import_collection.decoarsen_imported_list()
            if decoarsen_array is not None:
                log.info('static properties decoarsened')
                prop_import_collection.add_cached_array_to_imported_list(decoarsen_array,
                                                                         'decoarsen',
                                                                         'DECOARSEN',
                                                                         discrete = True,
                                                                         uom = None,
                                                                         time_index = None,
                                                                         null_value = -1,
                                                                         property_kind = 'discrete')
        log.info('writing ' + str(len(prop_import_collection.imported_list)) + ' properties to hdf5 file ' +
                 resqml_file_root + '.h5')
    elif not ensemble_case_dirs_root:
        log.info('no static grid properties to import')
        prop_import_collection = None
    grid.write_hdf5_from_caches(resqml_file_root + '.h5',
                                geometry = False,
                                imported_properties = prop_import_collection,
                                write_active = True)
    # remove cached static property arrays from memory
    if prop_import_collection is not None:
        prop_import_collection.remove_all_cached_arrays()

    ts_selection = None
    if using_vdb and vdb_recurrent_properties and timestep_selection is not None and str(timestep_selection) != 'none':
        if prop_import_collection is None:
            prop_import_collection = rp.GridPropertyCollection()
            prop_import_collection.set_grid(grid)
        # extract timestep dates from summary file (this info might be hidden in the recurrent binary files but I couldn't find it
        # todo: create cut down time series from recurrent files and differentiate between reporting time index and mapped time step number
        full_time_series = rts.time_series_from_nexus_summary(summary_file)
        if full_time_series is None:
            log.error('failed to fetch time series from Nexus summary file; recurrent data excluded')
            tm.log_nexus_tm('error')
        else:
            full_time_series.set_model(model)
            timestep_list = vdbase.grid_list_of_timesteps(
                grid_title)  # get list of timesteps for which recurrent files exist
            recur_time_series = None
            for timestep_number in timestep_list:
                if isinstance(timestep_selection, list):
                    if timestep_number not in timestep_selection:
                        continue
                else:
                    if timestep_selection == 'first':
                        if timestep_number != timestep_list[0]:
                            break
                    elif timestep_selection == 'last':
                        if timestep_number != timestep_list[-1]:
                            continue
                    elif timestep_selection == 'first and last':
                        if timestep_number != timestep_list[0] and timestep_number != timestep_list[-1]:
                            continue
                    # default to importing all timesteps
                stamp = full_time_series.timestamp(timestep_number)
                if stamp is None:
                    log.error('timestamp number for which recurrent data exists was not found in summary file: ' +
                              str(timestep_number))
                    continue
                recur_prop_list = vdbase.grid_list_of_recurrent_properties(grid_title, timestep_number)
                common_recur_prop_set = set()
                if recur_time_series is None:
                    recur_time_series = rts.TimeSeries(model, first_timestamp = stamp)
                    if recur_prop_list is not None:
                        common_recur_prop_set = set(recur_prop_list)
                else:
                    recur_time_series.add_timestamp(stamp)
                    if recur_prop_list is not None:
                        common_recur_prop_set = common_recur_prop_set.intersection(set(recur_prop_list))
                step_import_collection = rp.GridPropertyCollection()
                step_import_collection.set_grid(grid)
                # for each property for this timestep, cache array and add to recur prop import collection for this time step
                if recur_prop_list:
                    for keyword in recur_prop_list:
                        if not keyword or not keyword.isalnum():
                            continue
                        prop_kind, facet_type, facet = rp.property_kind_and_facet_from_keyword(keyword)
                        step_import_collection.import_vdb_recurrent_property_to_cache(
                            vdbase,
                            timestep_number,  # also used as time_index?
                            keyword,
                            grid_name = grid_title,
                            property_kind = prop_kind,
                            facet_type = facet_type,
                            facet = facet)
                # extend hdf5 with cached arrays for this timestep
                log.info('number of recurrent grid property arrays for timestep: ' + str(timestep_number) + ' is: ' +
                         str(step_import_collection.number_of_imports()))
                if decoarsen_array is not None:
                    log.info('decoarsening recurrent properties for timestep: ' + str(timestep_number))
                    step_import_collection.decoarsen_imported_list(decoarsen_array = decoarsen_array)
                log.info('extending hdf5 file with recurrent properties for timestep: ' + str(timestep_number))
                grid.write_hdf5_from_caches(resqml_file_root + '.h5',
                                            mode = 'a',
                                            geometry = False,
                                            imported_properties = step_import_collection,
                                            write_active = False)
                # add imported list for this timestep to full imported list
                prop_import_collection.inherit_imported_list_from_other_collection(step_import_collection)
                log.debug('total number of property arrays after timestep: ' + str(timestep_number) + ' is: ' +
                          str(prop_import_collection.number_of_imports()))
                # remove cached copies of arrays
                step_import_collection.remove_all_cached_arrays()

            ts_node = full_time_series.create_xml(title = 'simulator full timestep series')
            model.time_series = ts_node  # save as the primary time series for the model
            ts_uuid = rqet.uuid_for_part_root(ts_node)
            # create xml for recur_time_series (as well as for full_time_series) and add as part; not needed?
            if recur_time_series is not None:
                rts_node = recur_time_series.create_xml(title = 'simulator recurrent array timestep series')
                if use_compressed_time_series:
                    ts_uuid = rqet.uuid_for_part_root(rts_node)
                    ts_selection = timestep_list

    # add imported properties parts to model, building property parts list
    if prop_import_collection is not None and prop_import_collection.imported_list is not None:
        prop_import_collection.set_grid(grid)  # update to pick up on recently created xml root node for grid
        prop_import_collection.create_xml_for_imported_list_and_add_parts_to_model(
            ext_uuid, time_series_uuid = ts_uuid, selected_time_indices_list = ts_selection)
        if create_property_set:
            prop_import_collection.create_property_set_xml('property set for import for grid ' + str(grid_title))

    # mark model as modified (will already have happened anyway)
    model.set_modified()

    # create epc file
    log.info('storing model in epc file ' + resqml_file_root + '.epc')
    model.store_epc(resqml_file_root + '.epc')

    # return resqml model
    return model



"""A submodule containing functions relating to grid pillars."""

import logging

log = logging.getLogger(__name__)

import numpy as np
import resqpy.olio.vector_utilities as vec


def create_column_pillar_mapping(grid):
    """Creates an array attribute holding set of 4 pillar indices for each I, J column of cells.

    returns:
       numpy integer array of shape (nj, ni, 2, 2) where the last two indices are jp, ip;
       the array contains the pillar index for each of the 4 corners of each column of cells

    notes:
       the array is also cached as an attribute of the grid object: grid.pillars_for_column
       for grids with split coordinates lines (faults), this array allows for fast access to
       the correct pillar data for the corner of a column of cells;
       here and elsewhere, ip & jp (& kp) refer to a 0 or 1 index which determines the side
       of a cell, ip & jp together select one of the four corners of a column;
       the pillar index is a single integer, which is used as the second index into the points
       array for a grid geometry with split pillars;
       for unsplit grid geometries, such a pillar index must be converted back into a j', i'
       pair of indices (or the points array must be reshaped to combine the two indices into one)

    :meta common:
    """

    if hasattr(grid, 'pillars_for_column') and grid.pillars_for_column is not None:
        return grid.pillars_for_column

    grid.cache_all_geometry_arrays()

    grid.pillars_for_column = np.empty((grid.nj, grid.ni, 2, 2), dtype = int)
    ni_plus_1 = grid.ni + 1

    for j in range(grid.nj):
        grid.pillars_for_column[j, :, 0, 0] = np.arange(j * ni_plus_1, (j + 1) * ni_plus_1 - 1, dtype = int)
        grid.pillars_for_column[j, :, 0, 1] = np.arange(j * ni_plus_1 + 1, (j + 1) * ni_plus_1, dtype = int)
        grid.pillars_for_column[j, :, 1, 0] = np.arange((j + 1) * ni_plus_1, (j + 2) * ni_plus_1 - 1, dtype = int)
        grid.pillars_for_column[j, :, 1, 1] = np.arange((j + 1) * ni_plus_1 + 1, (j + 2) * ni_plus_1, dtype = int)

    if grid.has_split_coordinate_lines:
        unsplit_pillar_count = (grid.nj + 1) * ni_plus_1
        extras_count = len(grid.split_pillar_indices_cached)
        for extra_index in range(extras_count):
            primary = grid.split_pillar_indices_cached[extra_index]
            primary_ji0 = divmod(primary, grid.ni + 1)
            extra_pillar_index = unsplit_pillar_count + extra_index
            if extra_index == 0:
                start = 0
            else:
                start = grid.cols_for_split_pillars_cl[extra_index - 1]
            for cpscl_index in range(start, grid.cols_for_split_pillars_cl[extra_index]):
                col = grid.cols_for_split_pillars[cpscl_index]
                j, i = divmod(col, grid.ni)
                jp = primary_ji0[0] - j
                ip = primary_ji0[1] - i
                assert (jp == 0 or jp == 1) and (ip == 0 or ip == 1)
                grid.pillars_for_column[j, i, jp, ip] = extra_pillar_index

    return grid.pillars_for_column


def pillar_foursome(grid, ji0, none_if_unsplit = False):
    """Returns an int array of the natural pillar indices applicable to each column around primary.

    arguments:
       ji0 (pair of ints): the pillar indices (j0, i0) of the primary pillar of interest
       none_if_unsplit (boolean, default False): if True and the primary pillar is unsplit, None is returned; if False,
          a foursome is returned full of the natural index of the primary pillar

    returns:
       numpy int array of shape (2, 2) being the natural pillar indices (second axis index in raw points array)
       applicable to each of the four columns around the primary pillar; axes of foursome are (jp, ip); if the
       primary pillar is unsplit, None is returned if none_if_unsplit is set to True, otherwise the foursome as
       usual
    """

    j0, i0 = ji0

    grid.cache_all_geometry_arrays()

    primary = (grid.ni + 1) * j0 + i0
    foursome = np.full((2, 2), primary, dtype = int)  # axes are: jp, ip
    if not grid.has_split_coordinate_lines:
        return None if none_if_unsplit else foursome
    extras = np.where(grid.split_pillar_indices_cached == primary)[0]
    if len(extras) == 0:
        return None if none_if_unsplit else foursome

    primary_count = (grid.nj + 1) * (grid.ni + 1)
    assert len(grid.cols_for_split_pillars) == grid.cols_for_split_pillars_cl[-1]
    for cpscl_index in extras:
        if cpscl_index == 0:
            start_index = 0
        else:
            start_index = grid.cols_for_split_pillars_cl[cpscl_index - 1]
        for csp_index in range(start_index, grid.cols_for_split_pillars_cl[cpscl_index]):
            natural_col = grid.cols_for_split_pillars[csp_index]
            col_j0_e, col_i0_e = divmod(natural_col, grid.ni)
            col_j0_e -= (j0 - 1)
            col_i0_e -= (i0 - 1)
            assert col_j0_e in [0, 1] and col_i0_e in [0, 1]
            foursome[col_j0_e, col_i0_e] = primary_count + cpscl_index

    return foursome


def pillar_distances_sqr(grid, xy, ref_k0 = 0, kp = 0, horizon_points = None):
    """Returns array of the square of the distances of primary pillars in x,y plane to point xy.

    arguments:
       xy (float pair): the xy coordinate to compute the pillar distances to
       ref_k0 (int, default 0): the horizon layer number to use
       horizon_points (numpy array, optional): if present, should be array as returned by
          horizon_points() method; pass for efficiency in case of multiple calls
    """

    # note: currently works with unmasked data and using primary pillars only
    pe_j = grid.extent_kji[1] + 1
    pe_i = grid.extent_kji[2] + 1
    if horizon_points is None:
        horizon_points = grid.horizon_points(ref_k0 = ref_k0, kp = kp)
    pillar_xy = horizon_points[:, :, 0:2]
    dxy = pillar_xy - xy
    dxy2 = dxy * dxy
    return (dxy2[:, :, 0] + dxy2[:, :, 1]).reshape((pe_j, pe_i))


def nearest_pillar(grid, xy, ref_k0 = 0, kp = 0):
    """Returns the (j0, i0) indices of the primary pillar with point closest in x,y plane to point xy."""

    # note: currently works with unmasked data and using primary pillars only
    pe_i = grid.extent_kji[2] + 1
    sum_dxy2 = grid.pillar_distances_sqr(xy, ref_k0 = ref_k0, kp = kp)
    ji = np.nanargmin(sum_dxy2)
    j, i = divmod(ji, pe_i)
    return (j, i)


def nearest_rod(grid, xyz, projection, axis, ref_slice0 = 0, plus_face = False):
    """Returns the (k0, j0) or (k0 ,i0) indices of the closest point(s) to xyz(s); projection is 'xy', 'xz' or 'yz'.

    note:
       currently only for unsplit grids
    """

    x_sect = grid.unsplit_x_section_points(axis, ref_slice0 = ref_slice0, plus_face = plus_face)
    if type(xyz) is np.ndarray and xyz.ndim > 1:
        assert xyz.shape[-1] == 3
        result_shape = list(xyz.shape)
        result_shape[-1] = 2
        nearest = np.empty(tuple(result_shape), dtype = int).reshape((-1, 2))
        for i, p in enumerate(xyz.reshape((-1, 3))):
            nearest[i] = vec.nearest_point_projected(p, x_sect, projection)
        return nearest.reshape(tuple(result_shape))
    else:
        return vec.nearest_point_projected(xyz, x_sect, projection)



"""well_utils.py: functions used by the classes in resqpy.well"""

version = '10th November 2021'

# Nexus is a registered trademark of the Halliburton Company
# RMS and ROXAR are registered trademarks of Roxar Software Solutions AS, an Emerson company

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.olio.grid_functions as gf
import resqpy.olio.intersection as intersect
import resqpy.olio.keyword_files as kf
import resqpy.olio.xml_et as rqet


def load_hdf5_array(object, node, array_attribute, tag = 'Values', dtype = 'float', model = None):
    """Loads the property array data as an attribute of object, from the hdf5 referenced in xml node.

    :meta private:
    """

    assert (rqet.node_type(node) in ['DoubleHdf5Array', 'IntegerHdf5Array', 'Point3dHdf5Array'])
    if model is None:
        model = object.model
    h5_key_pair = model.h5_uuid_and_path_for_node(node, tag = tag)
    if h5_key_pair is None:
        return None
    return model.h5_array_element(h5_key_pair,
                                  index = None,
                                  cache_array = True,
                                  dtype = dtype,
                                  object = object,
                                  array_attribute = array_attribute)


def extract_xyz(xyz_node):
    """Extracts an x,y,z coordinate from a solitary point xml node.

    argument:
       xyz_node: the xml node representing the solitary point (in 3D space)

    returns:
       triple float: (x, y, z) coordinates as a tuple
    """

    if xyz_node is None:
        return None
    xyz = np.zeros(3)
    for axis in range(3):
        xyz[axis] = rqet.find_tag_float(xyz_node, 'Coordinate' + str(axis + 1), must_exist = True)
    return tuple(xyz)


def well_names_in_cellio_file(cellio_file):
    """Returns a list of well names as found in the RMS blocked well export cell I/O file."""

    well_list = []
    with open(cellio_file, 'r') as fp:
        while True:
            kf.skip_blank_lines_and_comments(fp)
            line = fp.readline()  # file format version number?
            if line == '':
                break  # end of file
            fp.readline()  # 'Undefined'
            words = fp.readline().split()
            assert len(words), 'missing header info (well name) in cell I/O file'
            well_list.append(words[0])
            while not kf.blank_line(fp):
                fp.readline()  # skip to block of data for next well
    return well_list


# 'private' functions


def find_entry_and_exit(cp, entry_vector, exit_vector, well_name):
    """Returns (entry_axis, entry_polarity, entry_xyz, exit_axis, exit_polarity, exit_xyz).

    :meta private:
    """

    cell_centre = np.mean(cp, axis = (0, 1, 2))
    face_triangles = gf.triangles_for_cell_faces(cp).reshape(-1, 3, 3)  # flattened first index 4 values per face
    entry_points = intersect.line_triangles_intersects(cell_centre, entry_vector, face_triangles, line_segment = True)
    entry_axis = entry_polarity = entry_xyz = exit_xyz = None
    for t in range(24):
        if not np.any(np.isnan(entry_points[t])):
            entry_xyz = entry_points[t]
            entry_axis = t // 8
            entry_polarity = (t - 8 * entry_axis) // 4
            break
    assert entry_axis is not None, 'failed to find entry face for a perforation in well ' + str(well_name)
    exit_points = intersect.line_triangles_intersects(cell_centre, exit_vector, face_triangles, line_segment = True)
    exit_axis = exit_polarity = None
    for t in range(24):
        if not np.any(np.isnan(exit_points[t])):
            exit_xyz = entry_points[t]
            exit_axis = t // 8
            exit_polarity = (t - 8 * exit_axis) // 4
            break
    assert exit_axis is not None, 'failed to find exit face for a perforation in well ' + str(well_name)

    return (entry_axis, entry_polarity, entry_xyz, exit_axis, exit_polarity, exit_xyz)


def _as_optional_array(arr):
    """If not None, cast as numpy array.

    Casting directly to an array can be problematic: np.array(None) creates an unsized array, which is potentially
    confusing.
    """
    if arr is None:
        return None
    else:
        return np.array(arr)


def _pl(i, e = False):
    return '' if i == 1 else 'es' if e else 's'


def _derive_from_wellspec_verify_col_list(add_properties):
    """ Verify additional properties to be added to the WELLSPEC file.

    argument:
       add_properties (boolean): if True, the additional properties specified will be added to the WELLSPEC file

    returns:
       list of columns to be added to the WELLSPEC file
    """

    if add_properties:
        if isinstance(add_properties, list):
            col_list = ['IW', 'JW', 'L'] + [col.upper() for col in add_properties if col not in ['IW', 'JW', 'L']]
        else:
            col_list = []
    else:
        col_list = ['IW', 'JW', 'L', 'ANGLA', 'ANGLV']
    return col_list


def _derive_from_wellspec_check_grid_name(check_grid_name, grid, col_list):
    """ Verify the grid object to which the cell indices in the WELLSPEC table belong.

    arguments:
       check_grid_name (boolean): if True, the citation title of the grid will be extracted and returned
       grid (grid object): the grid object whose citation titles will be returned
       col_list (list): list of strings of column names to be added to the WELLSPEC file. If a citation title is
                        extracted from the grid object, 'GRID' will be added to the col_list

    returns:
       string of grid citation title extracted from the grid object
       list of columns to be added to the WELLSPEC file
    """

    if check_grid_name:
        grid_name = rqet.citation_title_for_node(grid.root).upper()
        if not grid_name:
            name_for_check = None
        else:
            col_list.append('GRID')
            name_for_check = grid_name
    else:
        name_for_check = None
    return name_for_check, col_list



"""High level fault throw scaling functions."""

import logging

log = logging.getLogger(__name__)

import os
import math as maths
import numpy as np

import resqpy.fault as rqf
import resqpy.model as rq
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet

from resqpy.derived_model._common import _displacement_properties, _prepare_simple_inheritance, _write_grid, _establish_model_and_source_grid
from resqpy.derived_model._copy_grid import copy_grid


def fault_throw_scaling(epc_file,
                        source_grid = None,
                        scaling_factor = None,
                        connection_set = None,
                        scaling_dict = None,
                        ref_k0 = 0,
                        ref_k_faces = 'top',
                        cell_range = 0,
                        offset_decay = 0.5,
                        store_displacement = False,
                        inherit_properties = False,
                        inherit_realization = None,
                        inherit_all_realizations = False,
                        inherit_gcs = True,
                        new_grid_title = None,
                        new_epc_file = None):
    """Extends epc with a new grid with fault throws multiplied by scaling factors.

    arguments:
       epc_file (string): file name to rewrite the model's xml to; if source grid is None, model is loaded from this file
       source_grid (grid.Grid object, optional): if None, the epc_file is loaded and it should contain one ijk grid object
          (or one 'ROOT' grid) which is used as the source grid
       scaling_factor (float, optional): if present, the default scaling factor to apply to split pillars which do not
          appear in any of the faults in the scaling dictionary; if None, such pillars are left unchanged
       connection_set (fault.GridConnectionSet object): the connection set with associated fault feature list, used to
          identify which faces (and hence pillars) belong to which named fault
       scaling_dict (dictionary mapping string to float): the scaling factor to apply to each named fault; any faults not
          included in the dictionary will be left unadjusted (unless a default scaling factor is given as scaling_factor arg)
       ref_k0 (integer, default 0): the reference layer (zero based) to use when determining the pre-existing throws
       ref_k_faces (string, default 'top'): 'top' or 'base' identifying which bounding interface to use as the reference
       cell_range (integer, default 0): the number of cells away from faults which will have depths adjusted to spatially
          smooth the effect of the throw scaling (ie. reduce sudden changes in gradient due to the scaling)
       offset_decay (float, default 0.5): DEPRECATED; ignored
       store_displacement (boolean, default False): if True, 3 grid property parts are created, one each for x, y, & z
          displacement of cells' centres brought about by the fault throw scaling
       inherit_properties (boolean, default False): if True, the new grid will have a copy of any properties associated
          with the source grid
       inherit_realization (int, optional): realization number for which properties will be inherited; ignored if
          inherit_properties is False
       inherit_all_realizations (boolean, default False): if True (and inherit_realization is None), properties for all
          realizations will be inherited; if False, only properties with a realization of None are inherited; ignored if
          inherit_properties is False or inherit_realization is not None
       inherit_gcs (boolean, default True): if True, any grid connection set objects related to the source grid will be
          inherited by the modified grid
       new_grid_title (string): used as the citation title text for the new grid object
       new_epc_file (string, optional): if None, the source epc_file is extended with the new grid object; if present,
          a new epc file (& associated h5 file) is created to contain the derived grid (& crs)

    returns:
       new grid (grid.Grid object), with fault throws scaled according to values in the scaling dictionary

    notes:
       grid points are moved along pillar lines;
       stretch is towards or away from mid-point of throw;
       same shift is applied to all layers along pillar;
       pillar lines assumed to be straight;
       the offset decay argument might be changed in a future version to give improved smoothing;
       if a large fault is represented by a series of parallel minor faults 'stepping' down, each minor fault will have the
       scaling factor applied independently, leading to some unrealistic results
    """

    assert epc_file or new_epc_file, 'epc file name not specified'
    if new_epc_file and epc_file and (
        (new_epc_file == epc_file) or
        (os.path.exists(new_epc_file) and os.path.exists(epc_file) and os.path.samefile(new_epc_file, epc_file))):
        new_epc_file = None
    model, source_grid = _establish_model_and_source_grid(epc_file, source_grid)
    assert source_grid.grid_representation == 'IjkGrid'
    assert model is not None

    assert source_grid.has_split_coordinate_lines, 'cannot scale fault throws in unfaulted grid'
    assert scaling_factor is not None or (connection_set is not None and scaling_dict is not None)

    if ref_k_faces == 'base':
        ref_k0 += 1
    assert ref_k0 >= 0 and ref_k0 <= source_grid.nk, 'reference layer out of range'

    # take a copy of the grid
    log.debug('copying grid')
    grid = copy_grid(source_grid, model)
    grid.cache_all_geometry_arrays()  # probably already cached anyway

    # todo: handle pillars with no geometry defined, and cells without geometry defined
    assert grid.geometry_defined_for_all_pillars(), 'not all pillars have defined geometry'

    primaries = (grid.nj + 1) * (grid.ni + 1)
    offsets = np.zeros(grid.points_cached.shape[1:])

    if scaling_factor is not None:  # apply global scaling to throws
        _set_offsets_based_on_scaling_factor(grid, scaling_factor, offsets, ref_k0, primaries)

    if connection_set is not None and scaling_dict is not None:  # overwrite any global offsets with named fault throw adjustments
        _set_offsets_based_on_scaling_dict(grid, connection_set, scaling_dict, offsets, ref_k0)

    # initialise flag array for adjustments
    adjusted = np.zeros((primaries,), dtype = bool)

    # insert adjusted throws to all layers of split pillars
    grid.points_cached[:, grid.split_pillar_indices_cached, :] += offsets[grid.split_pillar_indices_cached, :].reshape(
        1, -1, 3)
    adjusted[grid.split_pillar_indices_cached] = True
    grid.points_cached[:, primaries:, :] += offsets[primaries:, :].reshape(1, -1, 3)

    # iteratively look for pillars neighbouring adjusted pillars, adjusting by a decayed amount
    adjusted = adjusted.reshape((grid.nj + 1, grid.ni + 1))
    while cell_range > 0:
        newly_adjusted = _neighbourly_adjustment(grid, offsets, adjusted, cell_range)
        adjusted = np.logical_or(adjusted, newly_adjusted)
        cell_range -= 1

    # check cell edge relative directions (in x,y) to ensure geometry is still coherent
    log.debug('checking grid geometry coherence')
    grid.check_top_and_base_cell_edge_directions()

    # build cell displacement property array(s)
    if store_displacement:
        log.debug('generating cell displacement property arrays')
        displacement_collection = _displacement_properties(grid, source_grid)
    else:
        displacement_collection = None

    collection = _prepare_simple_inheritance(grid, source_grid, inherit_properties, inherit_realization,
                                             inherit_all_realizations)
    if collection is None:
        collection = displacement_collection
    elif displacement_collection is not None:
        collection.inherit_imported_list_from_other_collection(displacement_collection, copy_cached_arrays = False)

    if new_grid_title is None or len(new_grid_title) == 0:
        new_grid_title = 'grid with fault throws scaled by ' + str(scaling_factor) + ' from ' +  \
                         str(rqet.citation_title_for_node(source_grid.root))

    gcs_list = []
    if inherit_gcs:
        gcs_uuids = model.uuids(obj_type = 'GridConnectionSetRepresentation', related_uuid = source_grid.uuid)
        for gcs_uuid in gcs_uuids:
            gcs = rqf.GridConnectionSet(model, uuid = gcs_uuid)
            gcs.cache_arrays()
            gcs_list.append((gcs, gcs.title))
        log.debug(f'{len(gcs_list)} grid connection sets to be inherited')

    # write model
    model.h5_release()
    if new_epc_file:
        _write_grid(new_epc_file, grid, property_collection = collection, grid_title = new_grid_title, mode = 'w')
        epc_file = new_epc_file
    else:
        ext_uuid, _ = model.h5_uuid_and_path_for_node(rqet.find_nested_tags(source_grid.root, ['Geometry', 'Points']),
                                                      'Coordinates')
        _write_grid(epc_file,
                    grid,
                    ext_uuid = ext_uuid,
                    property_collection = collection,
                    grid_title = new_grid_title,
                    mode = 'a')

    if len(gcs_list):
        log.debug(f'inheriting grid connection sets related to source grid: {source_grid.uuid}')
        _inherit_gcs_list(epc_file, gcs_list, source_grid, grid)

    return grid


def global_fault_throw_scaling(epc_file,
                               source_grid = None,
                               scaling_factor = None,
                               ref_k0 = 0,
                               ref_k_faces = 'top',
                               cell_range = 0,
                               offset_decay = 0.5,
                               store_displacement = False,
                               inherit_properties = False,
                               inherit_realization = None,
                               inherit_all_realizations = False,
                               inherit_gcs = True,
                               new_grid_title = None,
                               new_epc_file = None):
    """Rewrites epc with a new grid with all the fault throws multiplied by the same scaling factor.

    arguments:
       epc_file (string): file name to rewrite the model's xml to; if source grid is None, model is loaded from this file
       source_grid (grid.Grid object, optional): if None, the epc_file is loaded and it should contain one ijk grid object
          (or one 'ROOT' grid) which is used as the source grid
       scaling_factor (float): the scaling factor to apply to the throw across all split pillars
       ref_k0 (integer, default 0): the reference layer (zero based) to use when determining the pre-existing throws
       ref_k_faces (string, default 'top'): 'top' or 'base' identifying which bounding interface to use as the reference
       cell_range (integer, default 0): the number of cells away from faults which will have depths adjusted to spatially
          smooth the effect of the throw scaling (ie. reduce sudden changes in gradient due to the scaling)
       offset_decay (float, default 0.5): DEPRECATED; ignored
       store_displacement (boolean, default False): if True, 3 grid property parts are created, one each for x, y, & z
          displacement of cells' centres brought about by the fault throw scaling
       inherit_properties (boolean, default False): if True, the new grid will have a copy of any properties associated
          with the source grid
       inherit_realization (int, optional): realization number for which properties will be inherited; ignored if
          inherit_properties is False
       inherit_all_realizations (boolean, default False): if True (and inherit_realization is None), properties for all
          realizations will be inherited; if False, only properties with a realization of None are inherited; ignored if
          inherit_properties is False or inherit_realization is not None
       inherit_gcs (boolean, default True): if True, any grid connection set objects related to the source grid will be
          inherited by the modified grid
       new_grid_title (string): used as the citation title text for the new grid object
       new_epc_file (string, optional): if None, the source epc_file is extended with the new grid object; if present,
          a new epc file (& associated h5 file) is created to contain the derived grid (& crs)

    returns:
       new grid (grid.Grid object), with all fault throws scaled by the scaling factor

    notes:
       a scaling factor of 1 implies no change;
       calls fault_throw_scaling(), see also documentation for that function
    """

    return fault_throw_scaling(epc_file,
                               source_grid = source_grid,
                               scaling_factor = scaling_factor,
                               connection_set = None,
                               scaling_dict = None,
                               ref_k0 = ref_k0,
                               ref_k_faces = ref_k_faces,
                               cell_range = cell_range,
                               store_displacement = store_displacement,
                               inherit_properties = inherit_properties,
                               inherit_realization = inherit_realization,
                               inherit_all_realizations = inherit_all_realizations,
                               inherit_gcs = inherit_gcs,
                               new_grid_title = new_grid_title,
                               new_epc_file = new_epc_file)


def _set_offsets_based_on_scaling_factor(grid, scaling_factor, offsets, ref_k0, primaries):
    # fetch unsplit equivalent of grid points for reference layer interface
    log.debug('fetching unsplit equivalent grid points')
    unsplit_points = grid.unsplit_points_ref().reshape(grid.nk + 1, -1, 3)
    # determine existing throws on split pillars
    semi_throws = np.zeros(grid.points_cached.shape[1:])  # same throw applied to all layers
    unique_spi = np.unique(grid.split_pillar_indices_cached)
    semi_throws[unique_spi, :] = (grid.points_cached[ref_k0, unique_spi, :] - unsplit_points[ref_k0, unique_spi, :])
    semi_throws[primaries:, :] = (grid.points_cached[ref_k0, primaries:, :] -
                                  unsplit_points[ref_k0, grid.split_pillar_indices_cached, :]
                                 )  # unsplit points are mid points
    # ensure no adjustment in pillar where geometry is not defined in reference layer
    all_good = grid.geometry_defined_for_all_cells()
    if not all_good:
        log.warning('not all cells have defined geometry')
        semi_throws[:, :] = np.where(np.isnan(semi_throws), 0.0, semi_throws)
    # apply global scaling to throws
    offsets[:] = semi_throws * (scaling_factor - 1.0)


def _set_offsets_based_on_scaling_dict(grid, connection_set, scaling_dict, offsets, ref_k0):
    connection_set.cache_arrays()
    for fault_index in range(len(connection_set.feature_list)):
        fault_name = connection_set.fault_name_for_feature_index(fault_index)
        if fault_name not in scaling_dict:
            continue  # no scaling for this fault
        fault_scaling = scaling_dict[fault_name]
        if fault_scaling == 1.0:
            continue
        log.info('scaling throw on fault ' + fault_name + ' by factor of: {0:.4f}'.format(fault_scaling))
        kelp_j, kelp_i = connection_set.simplified_sets_of_kelp_for_feature_index(fault_index)
        p_list = []  # list of adjusted pillars
        for kelp in kelp_j:
            for ip in [0, 1]:
                p_a = grid.pillars_for_column[kelp[0], kelp[1], 1, ip]
                p_b = grid.pillars_for_column[kelp[0] + 1, kelp[1], 0, ip]  # other side of fault
                mid_point = 0.5 * (grid.points_cached[ref_k0, p_a] + grid.points_cached[ref_k0, p_b])
                if np.any(np.isnan(mid_point)):
                    continue
                if p_a not in p_list:
                    offsets[p_a] = (grid.points_cached[ref_k0, p_a] - mid_point) * (fault_scaling - 1.0)
                    p_list.append(p_a)
                if p_b not in p_list:
                    offsets[p_b] = (grid.points_cached[ref_k0, p_b] - mid_point) * (fault_scaling - 1.0)
                    p_list.append(p_b)
        for kelp in kelp_i:
            for jp in [0, 1]:
                p_a = grid.pillars_for_column[kelp[0], kelp[1], jp, 1]
                p_b = grid.pillars_for_column[kelp[0], kelp[1] + 1, jp, 0]  # other side of fault
                mid_point = 0.5 * (grid.points_cached[ref_k0, p_a] + grid.points_cached[ref_k0, p_b])
                if np.any(np.isnan(mid_point)):
                    continue
                if p_a not in p_list:
                    offsets[p_a] = (grid.points_cached[ref_k0, p_a] - mid_point) * (fault_scaling - 1.0)
                    p_list.append(p_a)
                if p_b not in p_list:
                    offsets[p_b] = (grid.points_cached[ref_k0, p_b] - mid_point) * (fault_scaling - 1.0)
                    p_list.append(p_b)


def _neighbourly_adjustment(grid, offsets, adjusted, cell_range):
    offset_decay = (maths.pow(2.0, cell_range) - 1.0) / (maths.pow(2.0, cell_range + 1) - 1.0)
    newly_adjusted = np.zeros((grid.nj + 1, grid.ni + 1), dtype = bool)
    for j in range(grid.nj + 1):
        for i in range(grid.ni + 1):
            if adjusted[j, i]:
                continue
            p = j * (grid.ni + 1) + i
            if p in grid.split_pillar_indices_cached:
                continue
            contributions = 0
            accum = 0.0
            if (i > 0) and adjusted[j, i - 1]:
                if j > 0:
                    accum += offsets[grid.pillars_for_column[j - 1, i - 1, 1, 0], 2]
                    contributions += 1
                if j < grid.nj:
                    accum += offsets[grid.pillars_for_column[j, i - 1, 0, 0], 2]
                    contributions += 1
            if (j > 0) and adjusted[j - 1, i]:
                if i > 0:
                    accum += offsets[grid.pillars_for_column[j - 1, i - 1, 0, 1], 2]
                    contributions += 1
                if i < grid.ni:
                    accum += offsets[grid.pillars_for_column[j - 1, i, 0, 0], 2]
                    contributions += 1
            if (i < grid.ni) and adjusted[j, i + 1]:
                if j > 0:
                    accum += offsets[grid.pillars_for_column[j - 1, i, 1, 1], 2]
                    contributions += 1
                if j < grid.nj:
                    accum += offsets[grid.pillars_for_column[j, i, 0, 1], 2]
                    contributions += 1
            if (j < grid.nj) and adjusted[j + 1, i]:
                if i > 0:
                    accum += offsets[grid.pillars_for_column[j, i - 1, 1, 1], 2]
                    contributions += 1
                if i < grid.ni:
                    accum += offsets[grid.pillars_for_column[j, i, 1, 0], 2]
                    contributions += 1
            if contributions == 0:
                continue
            dxy_dz = ((grid.points_cached[grid.nk, p, :2] - grid.points_cached[0, p, :2]) /
                      (grid.points_cached[grid.nk, p, 2] - grid.points_cached[0, p, 2]))
            offsets[p, 2] = offset_decay * accum / float(contributions)
            offsets[p, :2] = offsets[p, 2] * dxy_dz
            grid.points_cached[:, p, :] += offsets[p, :].reshape((1, 3))
            newly_adjusted[j, i] = True
    return newly_adjusted


def _inherit_gcs_list(epc_file, gcs_list, source_grid, grid):
    gcs_inheritance_model = rq.Model(epc_file)
    for gcs, gcs_title in gcs_list:
        # log.debug(f'inheriting gcs: {gcs_title}; old gcs uuid: {gcs.uuid}')
        gcs.uuid = bu.new_uuid()
        grid_list_modifications = []
        for gi, g in enumerate(gcs.grid_list):
            # log.debug(f'gcs uses grid: {g.title}; grid uuid: {g.uuid}')
            if bu.matching_uuids(g.uuid, source_grid.uuid):
                grid_list_modifications.append(gi)
        assert len(grid_list_modifications)
        for gi in grid_list_modifications:
            gcs.grid_list[gi] = grid
        gcs.model = gcs_inheritance_model
        gcs.write_hdf5()
        gcs.create_xml(title = gcs_title)
    gcs_inheritance_model.store_epc()
    gcs_inheritance_model.h5_release()
